"""
This is a script for the backend of the AAclustPlot object for all plotting functions.
"""
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.decomposition import PCA
from scipy.cluster.hierarchy import linkage, leaves_list

import aaanalysis.utils as ut
from ._utils_aaclust import _compute_medoids, _compute_centers

# I Helper Functions
# Computation helper functions
def _get_mean_rank(data):
    _df = data.copy()
    _df['BIC_rank'] = _df[ut.COL_BIC].rank(ascending=False)
    _df['CH_rank'] = _df[ut.COL_CH].rank(ascending=False)
    _df['SC_rank'] = _df[ut.COL_SC].rank(ascending=False)
    rank = _df[['BIC_rank', 'CH_rank', 'SC_rank']].mean(axis=1).round(2)
    return rank


def _get_components(X, model_class=None, n_components=2, model_kwargs=None):
    """
    Apply dimensionality reduction on X using the specified mode.

    Returns:
    df_components : DataFrame, shape (n_samples, n_components)
        The reduced components labeled "PC1", "PC2", etc.
    """
    model_kwargs["n_components"] = n_components
    # Initialize and fit the model
    model = model_class(**model_kwargs)
    model_name = model_class.__name__
    components = model.fit_transform(X)
    # Create DataFrame
    columns = [f"{model_name}{i + 1}" for i in range(n_components)]
    if isinstance(model, PCA):
        explained_var = 100 * model.explained_variance_ratio_
        columns = [f"{col.replace('PCA', 'PC')} ({explained_var[i]:.1f}%)" for i, col in enumerate(columns)]
    df_components = pd.DataFrame(components, columns=columns)
    return df_components, model


def _get_clustered_order(df, method='average'):
    linked = linkage(df.corr().fillna(0), method=method)
    new_order = leaves_list(linked)
    return df.iloc[:, new_order]


# II Main Functions
def plot_eval(df_eval=None, dict_xlims=None, figsize=None, colors=None):
    """Plot evaluation of AAclust clustering results"""
    df_eval[ut.COL_RANK] = _get_mean_rank(df_eval)
    df_eval = df_eval.sort_values(by=ut.COL_RANK, ascending=True)
    # Plotting
    fig, axes = plt.subplots(1, 4, sharey=True, figsize=figsize)
    for i, col in enumerate(ut.COLS_EVAL_AACLUST):
        ax = axes[i]
        sns.barplot(ax=ax, data=df_eval, y=ut.COL_NAME, x=col, color=colors[i])
        # Customize subplots
        ax.set_ylabel("")
        ax.set_xlabel(col)
        # Adjust spines
        ax = ut.adjust_spine_to_middle(ax=ax)
        if i == 0:
            ax.set_title("Clustering", weight="bold")
        elif i == 2:
            ax.set_title("Quality measures", weight="bold")
        ax.tick_params(axis='y', which='both', left=False)
        if i != 0:
            ut.x_ticks_0(ax=ax)
    # Set xlims
    if dict_xlims is not None:
        for i in dict_xlims:
            axes[i].set_xlim(dict_xlims[i])
    plt.tight_layout()
    plt.subplots_adjust(wspace=0.25, hspace=0)
    return fig, axes


@ut.catch_runtime_warnings()
def plot_center_or_medoid(X=None, labels=None,
                          plot_centers=True, metric="correlation",
                          component_x=1, component_y=1,
                          model_class=None, model_kwargs=None,
                          ax=None, figsize=(7, 6),
                          dot_alpha=0.75, dot_size=100,
                          legend=True, palette=None):
    """Plot compressed (e.g., by PCA) clustering results with highlighting cluster centers or cluster medoids"""
    n_components = max(component_x, component_y)
    df_components, model = _get_components(X, model_class, n_components, model_kwargs)
    # Get centers or medoids and obtain compress them
    if plot_centers:
        X_ref, labels_ref = _compute_centers(X, labels=labels)
    else:
        X_ref, labels_ref, _ = _compute_medoids(X, labels=labels, metric=metric)
    X_ref_transformed = model.transform(X_ref)
    # Plotting
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    n_clusters = len(set(labels))
    if palette is None:
        palette = sns.color_palette("husl", n_clusters)
    ax = sns.scatterplot(x=df_components.iloc[:, component_x - 1],
                         y=df_components.iloc[:, component_y - 1],
                         hue=labels, palette=palette, alpha=dot_alpha,
                         s=dot_size, ax=ax, legend=legend)
    # Highlight centers or medoids
    kwargs = dict(x=X_ref_transformed[:, component_x - 1],
                  y=X_ref_transformed[:, component_y - 1],
                  hue=labels_ref, palette=palette,
                  s=dot_size * 1.5, edgecolor="k", linewidth=1,
                  legend=False, ax=ax)
    if plot_centers:
        ax = sns.scatterplot(**kwargs, marker="X")
    else:
        ax = sns.scatterplot(**kwargs)
    plt.xlabel(df_components.columns[component_x - 1])
    plt.ylabel(df_components.columns[component_y - 1])
    if legend:
        ax.legend(title="clusters", bbox_to_anchor=(1, 0.95), loc='upper left')
    plt.tight_layout()
    return ax, df_components


def plot_correlation(df_corr=None, labels=None, labels_ref=None, cluster_x=True, method="average",
                     bar_position="left", bar_colors="gray",
                     bar_width_x=0.1, bar_spacing_x=0.1, bar_width_y=0.1, bar_spacing_y=0.1,
                     xtick_label_rotation=90, ytick_label_rotation=0,
                     vmin=-1, vmax=1, cmap="viridis", kwargs_heatmap=None):
    """Plots heatmap for clustering results with rows (y-axis) corresponding to scales and columns (x-axis) to clusters."""
    # Adjust order of df_corr
    if cluster_x:
        df_corr = _get_clustered_order(df_corr, method=method)
    # Plot heatmap
    _kwargs_heatmap = {"cmap": cmap, "vmin": vmin, "vmax": vmax,
                       "cbar_kws": {"label": "Pearson correlation"}}
    if kwargs_heatmap is not None:
        _kwargs_heatmap.update(kwargs_heatmap)
    ax = sns.heatmap(data=df_corr, **_kwargs_heatmap)
    # Adjust ticks
    ax.set_xticklabels(ax.get_xticklabels(), rotation=xtick_label_rotation, ha="right")
    ax.set_yticklabels(ax.get_yticklabels(), rotation=ytick_label_rotation)
    # Customizing colorbart tick lines
    if 'cbar' not in _kwargs_heatmap:
        cbar = ax.collections[0].colorbar
        lw = ut.plot_gco(option="axes.linewidth")
        fs = ut.plot_gco(option="font.size")
        cbar.ax.tick_params(axis='y', width=lw, length=6, color='black', labelsize=fs-1)
    # Add bars for highlighting clustering
    labels_are_sorted = list(labels) == sorted(labels)
    y_labels = labels
    x_labels = labels if labels_ref is None else labels_ref
    if labels_are_sorted:
        x_labels = list(sorted(x_labels))
    if bar_position is not None:
        for pos in bar_position:
            _labels = x_labels if pos in ["top", "bottom"] else y_labels
            _bar_width = bar_width_x if pos in ["top", "bottom"] else bar_width_y
            _bar_spacing = bar_spacing_x if pos in ["top", "bottom"] else bar_spacing_y
            ut.plot_add_bars(ax=ax, labels=_labels, bar_position=pos, set_tick_labels=True,
                             bar_spacing=_bar_spacing, bar_width=_bar_width, colors=bar_colors,
                             xtick_label_rotation=xtick_label_rotation,
                             ytick_label_rotation=ytick_label_rotation)
    plt.tight_layout()
    return ax


"""
This is a script for filtering scales from df_scales, such as using correlation or subcategory coverage.
"""
import time
import pandas as pd
import numpy as np

import aaanalysis.utils as ut

LIST_DATASETS_WITH_FEATURES = ["DOM_GSEC"]
FOLDER_FEATURES = ut.FOLDER_DATA + "features" + ut.SEP

# I Helper Functions
def check_name(name=None):
    """Check provided names of dataset"""
    if name not in LIST_DATASETS_WITH_FEATURES:
        raise ValueError(f"'name' should be one of the following: {LIST_DATASETS_WITH_FEATURES}")


# II Main Functions
def load_features(name="DOM_GSEC"):
    """
    Load feature sets for protein benchmarking datasets.

    Features are only provided for in-depth analyzed datasets available from the :func:`load_dataset`
    function. These are as follows:

        - 'DOM_GSEC' ([Breimann24c]_)

    Parameters
    ----------
    name : str, default='DOM_GSEC'
        The name of the dataset for which features are loaded.

    Returns
    -------
    df_feat : pd.DataFrame, shape (n_features, n_feature_info)
        A DataFrame with features and their statistical measures as provided :meth:`CPP.run` method,
        including their feature importance.

    Examples
    --------
    .. include:: examples/load_features.rst
    """
    # Check input
    check_name(name=name)
    # Load features
    df_feat = ut.read_excel_cached(FOLDER_FEATURES + f"FEATURES_{name}.xlsx")
    return df_feat



"""
This is a script for the backend of the plotting module functions used by other AAanalysis modules.
"""
import seaborn as sns
import matplotlib as mpl
from matplotlib import pyplot as plt
import matplotlib.lines as mlines
import warnings

from .check_type import check_number_range

# I Helper function
def _create_marker(color, label, marker, marker_size, lw, edgecolor, linestyle, hatch, hatchcolor):
    """Create custom marker based on input."""
    # Default marker (matching to plot)
    if marker is None:
        return mpl.patches.Patch(facecolor=color,
                                 label=label,
                                 lw=lw,
                                 hatch=hatch,
                                 edgecolor=hatchcolor)
    # If marker is '-', treat it as a line
    if marker == "-":
         return plt.Line2D(xdata=[0, 1], ydata=[0, 1],
                           color=color,
                           linestyle=linestyle,
                           lw=lw,
                           label=label)
    # Creates marker element without line (lw=0)
    return plt.Line2D(xdata=[0], ydata=[0],
                      marker=marker,
                      label=label,
                      markerfacecolor=color,
                      color=edgecolor,
                      markersize=marker_size,
                      lw=0,
                      markeredgewidth=lw)


# Check functions
def _marker_has(marker, val=None):
    if isinstance(marker, str):
        return marker == val
    elif marker is None:
        return False
    elif isinstance(marker, list):
        return any([x == val for x in marker])
    else:
        raise ValueError(f"'marker' ({marker}) is wrong")

def _marker_has_no(marker, val=None):
    if isinstance(marker, str):
        return marker != val
    elif marker is None:
        return False
    elif isinstance(marker, list):
        return any([x != val for x in marker])
    else:
        raise ValueError(f"'marker' ({marker}) is wrong")


# Checking functions for list inputs
def _check_list_cat(dict_color=None, list_cat=None):
    """Ensure items in list_cat are keys in dict_color and match in length."""
    if not list_cat:
        return list(dict_color.keys())
    if not all(elem in dict_color for elem in list_cat):
        missing_keys = [elem for elem in list_cat if elem not in dict_color]
        raise ValueError(f"The following keys in 'list_cat' are not in 'dict_colors': {', '.join(missing_keys)}")
    if len(dict_color) < len(list_cat):
        raise ValueError(
            f"'dict_colors' (n={len(dict_color)}) must contain >= elements than 'list_cat' (n={len(list_cat)}).")
    return list_cat


def _check_labels(list_cat=None, labels=None):
    """Validate labels and match their length to list_cat."""
    if labels is None:
        labels = list_cat
    if len(list_cat) != len(labels):
        raise ValueError(f"Length must match of 'labels' ({len(labels)}) and categories ({len(list_cat)}).")
    return labels


# Checking functions for inputs that can be list or single values (redundancy accepted for better user communication)
def _check_hatches(marker=None, hatch=None, list_cat=None):
    """Check validity of list_hatche."""
    valid_hatches = ['/', '\\', '|', '-', '+', 'x', 'o', 'O', '.', '*']
    # Check if hatch is valid
    if isinstance(hatch, str):
        if hatch not in valid_hatches:
            raise ValueError(f"'hatch' ('{hatch}') must be one of following: {valid_hatches}")
    if isinstance(hatch, list):
        wrong_hatch = [x for x in hatch if x not in valid_hatches]
        if len(wrong_hatch) != 0:
            raise ValueError(
                f"'hatch' contains wrong values ('{wrong_hatch}')! Should be one of following: {valid_hatches}")
        if len(hatch) != len(list_cat):
            raise ValueError(f"Length must match of 'hatch' ({hatch}) and categories ({list_cat}).")  # Check if hatch can be chosen
    # Warn for parameter conflicts
    if _marker_has_no(marker, val=None) and hatch:
        warnings.warn(f"'hatch' can only be applied to the default marker, set 'marker=None'.", UserWarning)
    # Create hatch list
    list_hatch = [hatch] * len(list_cat) if not isinstance(hatch, list) else hatch
    return list_hatch


def _check_marker(marker=None, list_cat=None, lw=0):
    """Check validity of markers"""
    # Add '-' for line and None for default marker
    valid_markers = [None, "-"] + list(mlines.Line2D.markers.keys())
    # Check if marker is valid
    if not isinstance(marker, list) and marker not in valid_markers:
        raise ValueError(f"'marker' ('{marker}') must be one of following: {valid_markers}")
    if isinstance(marker, list):
        wrong_markers = [x for x in marker if x not in valid_markers]
        if len(wrong_markers) != 0:
            raise ValueError(f"'marker' contains wrong values  ('{wrong_markers}'). Should be one of following: {valid_markers}")
        if len(marker) != len(list_cat):
            raise ValueError(f"Length must match of 'marker' ({marker}) and categories ({list_cat}).")
    # Warn for parameter conflicts
    if _marker_has(marker, val="-") and lw <= 0:
        warnings.warn(f"Marker lines ('-') are only shown if 'lw' ({lw}) is > 0.", UserWarning)
    # Create marker list
    list_marker = [marker] * len(list_cat) if not isinstance(marker, list) else marker
    return list_marker


def _check_marker_size(marker_size=10, list_cat=None):
    """Check size of markers"""
    # Check if marker_size is valid
    if isinstance(marker_size, (int, float)):
        check_number_range(name='marker_size', val=marker_size, min_val=0, accept_none=True, just_int=False)
    elif isinstance(marker_size, list):
        for i in marker_size:
            check_number_range(name='marker_size', val=i, min_val=0, accept_none=True, just_int=False)
    elif isinstance(marker_size, list) and len(marker_size) != len(list_cat):
        raise ValueError(f"Length must match of 'marker_size' (marker_size) and categories ({list_cat}).")
    else:
        raise ValueError(f"'marker_size' has wrong data type: {type(marker_size)}")
    # Create marker_size list
    list_marker_size = [marker_size] * len(list_cat) if not isinstance(marker_size, list) else marker_size
    return list_marker_size


def _check_linestyle(linestyle=None, list_cat=None, marker=None):
    """Check validity of linestyle."""
    _lines = ['-', '--', '-.', ':', ]
    _names = ["solid", "dashed", "dashed-doted", "dotted"]
    valid_mls = _lines + _names
    # Check if marker_linestyle is valid
    if isinstance(linestyle, list):
        wrong_mls = [x for x in linestyle if x not in valid_mls]
        if len(wrong_mls) != 0:
            raise ValueError(
                f"'marker_linestyle' contains wrong values ('{wrong_mls}')! Should be one of following: {valid_mls}")
        if len(linestyle) != len(list_cat):
            raise ValueError(f"Length must match of 'marker_linestyle' ({linestyle}) and categories ({list_cat}).")
    # Check if marker_linestyle is conflicting with other settings
    if isinstance(linestyle, str):
        if linestyle not in valid_mls:
            raise ValueError(f"'marker_linestyle' ('{linestyle}') must be one of following: {_lines},"
                             f" or corresponding names: {_names} ")
    # Warn for parameter conflicts
    if linestyle is not None and _marker_has_no(marker, val="-"):
        warnings.warn(f"'linestyle' ({linestyle}) is only applicable to marker lines ('-'), not to '{marker}'.", UserWarning)
    # Create list_marker_linestyle list
    list_marker_linestyle = [linestyle] * len(list_cat) if not isinstance(linestyle, list) else linestyle
    return list_marker_linestyle


# II Main Functions
# DEV: General function for plot_gcfs
def plot_gco(option='font.size', show_options=False):
    """Get current option from plotting context"""
    current_context = sns.plotting_context()
    if show_options:
        print(current_context)
    try:
        option_value = current_context[option]  # Typically font_size
    except KeyError:
        options = list(current_context.keys())
        raise ValueError(f"Option not valid, select from the following: {options}")
    return option_value

# DEV: plot_get_cdict and plot_get_cmap are implemented in main utils

# Remaining backend plotting functions
def plot_get_clist_(n_colors=3):
    """Get manually curated list of 2 to 9 colors."""
    # Base lists
    list_colors_3_to_4 = ["tab:gray", "tab:blue", "tab:red", "tab:orange"]
    list_colors_5_to_6 = ["tab:blue", "tab:cyan", "tab:gray","tab:red",
                          "tab:orange", "tab:brown"]
    list_colors_8_to_9 = ["tab:blue", "tab:orange", "tab:green", "tab:red",
                          "tab:gray", "gold", "tab:cyan", "tab:brown",
                          "tab:purple"]
    # Two classes
    if n_colors == 2:
        return ["tab:blue", "tab:red"]
    # Control/base + 2-3 classes
    elif n_colors in [3, 4]:
        return list_colors_3_to_4[0:n_colors]
    # 5-7 classes (gray in middle as visual "breather")
    elif n_colors in [5, 6]:
        return list_colors_5_to_6[0:n_colors]
    elif n_colors == 7:
        return ["tab:blue", "tab:cyan", "tab:purple", "tab:gray",
                "tab:red", "tab:orange", "tab:brown"]
    # 8-9 classes (colors from scale categories)
    elif n_colors in [8, 9]:
        return list_colors_8_to_9[0:n_colors]
    else:
        return sns.color_palette(palette="husl", n_colors=n_colors)


def plot_legend_(ax=None, dict_color=None, list_cat=None, labels=None,
                 loc="upper left", loc_out=False, y=None, x=None, n_cols=3,
                 labelspacing=0.2, columnspacing=1.0, handletextpad=0.8, handlelength=2.0,
                 fontsize=None, fontsize_title=None, weight_font="normal", weight_title="normal",
                 marker=None, marker_size=10, lw=0, linestyle=None, edgecolor=None,
                 hatch=None, hatchcolor="white", title=None, title_align_left=True,
                 frameon=False, **kwargs):
    """Sets an independently customizable plot legend"""
    # Check input
    if ax is None:
        ax = plt.gca()
    list_cat = _check_list_cat(dict_color=dict_color, list_cat=list_cat)
    labels = _check_labels(list_cat=list_cat, labels=labels)
    marker = _check_marker(marker=marker, list_cat=list_cat, lw=lw)
    hatch = _check_hatches(marker=marker, hatch=hatch, list_cat=list_cat)
    linestyle = _check_linestyle(linestyle=linestyle, list_cat=list_cat, marker=marker)
    marker_size = _check_marker_size(marker_size=marker_size, list_cat=list_cat)
    # Remove existing legend
    if ax.get_legend() is not None and len(ax.get_legend().get_lines()) > 0:
        ax.legend_.remove()
    # Update legend arguments
    args = dict(loc=loc, ncol=n_cols, fontsize=fontsize, labelspacing=labelspacing, columnspacing=columnspacing,
                handletextpad=handletextpad, handlelength=handlelength, borderpad=0, title=title,
                edgecolor=edgecolor, prop={"weight": weight_font, "size": fontsize}, frameon=frameon)
    args.update(kwargs)
    if fontsize_title:
        args["title_fontproperties"] = {"weight": weight_title, "size": fontsize_title}
    if loc_out:
        x, y = x or 0, y or -0.25
    if x or y:
        args["bbox_to_anchor"] = (x or 0, y or 1)
    # Create handles and legend
    handles = [_create_marker(dict_color[cat], labels[i], marker[i], marker_size[i],
                              lw, edgecolor, linestyle[i], hatch[i], hatchcolor)
               for i, cat in enumerate(list_cat)]
    legend = ax.legend(handles=handles, labels=labels, **args)
    if title_align_left:
        legend._legend_box.align = "left"
    return ax



"""
This is a script for the backend of the CPPPlot.feature_map method.
"""
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import warnings

import aaanalysis.utils as ut

from ._utils_cpp_plot_elements import PlotElements
from ._utils_cpp_plot_positions import PlotPartPositions
from ._utils_cpp_plot_map import plot_heatmap_

# I Helper functions
def _get_sorted_list_cat(df_feat=None, col_cat=None):
    """Case non-sensitive sorted list of scale categories"""
    df = df_feat.copy()
    # Sort case non-sensitive
    df = df.sort_values(by=[ut.COL_CAT, ut.COL_SUBCAT], key=lambda x: x.str.lower())
    list_cat = list(df[col_cat].drop_duplicates())
    return list_cat


# Add feature importance plot elements
def plot_feat_importance_bars(ax=None, df_feat=None, col_cat=None, col_imp=None,
                              annotation_th=None, label=None,
                              fontsize_label=12,
                              fontsize_annotations=11,
                              fontsize_imp_bar=9,
                              weight_annotation="bold"):
    """Display a horizontal bar plot for feature importance sorted by categories."""
    # Get feature importance per scale class
    df_imp = df_feat[[col_cat, col_imp]].groupby(by=col_cat).sum()
    dict_imp = dict(zip(df_imp.index, df_imp[col_imp]))
    list_cat = _get_sorted_list_cat(df_feat=df_feat, col_cat=col_cat)
    list_imp = [dict_imp[x] for x in list_cat]

    # Maximum value

    # Plot bars
    plt.sca(ax)
    ax.barh(list_cat, list_imp, color=ut.COLOR_FEAT_IMP, edgecolor="white", align="edge")
    ax.set_xlabel(label, size=fontsize_label, weight="bold", ha="right", position=(1, 0))
    ax.xaxis.set_label_position('top')

    # Add annotations
    v_max = int(np.ceil(max(list_imp)))
    annotation_th = v_max / 2 if annotation_th is None else annotation_th
    for i, val in enumerate(list_imp):
        if val >= annotation_th:
            plt.text(val, i + 0.45, f"{round(val, 1)}% ",
                     va="center", ha="right",
                     weight=weight_annotation,
                     color="white",
                     size=fontsize_imp_bar)

    # Adjust ticks
    ax.tick_params(axis='y', which='both', length=0, labelsize=0)
    ax.tick_params(axis='x', which='both', labelsize=fontsize_annotations, pad=-1)
    sns.despine(ax=ax, bottom=True, top=False)
    plt.xlim(0, v_max)
    ut.x_ticks_0(ax)

    # Adjust plot size
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        plt.tight_layout()


def add_feat_importance_map(ax=None, df_feat=None, df_cat=None,
                            col_cat=None, col_imp=None,
                            imp_ths=(0.2, 0.5, 1),
                            imp_marker_size=(3, 5.5, 8),
                            start=None, args_len=None):
    """Overlay feature importance symbols on the heatmap based on the importance values."""
    th1, th2, th3 = imp_ths
    ms1, ms2, ms3 = imp_marker_size
    pp = PlotPartPositions(**args_len, start=start)
    df_pos = pp.get_df_pos(df_feat=df_feat.copy(), df_cat=df_cat,
                           col_cat=col_cat, col_val=col_imp,
                           value_type="sum", normalize=False)
    _df = pd.melt(df_pos.reset_index(), id_vars="index")
    _df.columns = [ut.COL_SUBCAT, ut.COL_POSITION, col_imp]
    _list_sub_cat = _df[ut.COL_SUBCAT].unique()
    for i, sub_cat in enumerate(_list_sub_cat):
        _dff = _df[_df[ut.COL_SUBCAT] == sub_cat]
        for pos, val in enumerate(_dff[col_imp]):
            _symbol = "■"
            color = "black"
            size = ms3 if val >= th3 else (ms2 if val >= th2 else ms1)
            _args_symbol = dict(ha="center", va="center", color=color, size=size)
            if val >= th1:
                ax.text(pos + 0.5, i + 0.5, _symbol, **_args_symbol)


def add_feat_importance_legend(ax=None,
                               legend_imp_xy=None,
                               imp_ths=(0.2, 0.5, 1),
                               label=None,
                               fontsize_title=None,
                               fontsize_annotations=None):
    """Add a custom legend indicating the meaning of feature importance symbols."""
    # Define the sizes for the legend markers
    list_labels = [f"  >{float(x)}%" for x in imp_ths]
    list_imp_marker_sizess = [3, 5, 7]

    # Create the legend handles manually
    legend_handles = [
        plt.Line2D([0], [0], marker='s', color='w', label=label,
                   markersize=size, markerfacecolor='black', linewidth=0)
        for label, size in zip(list_labels, list_imp_marker_sizess)]

    # Create the second legend
    ax.legend(handles=legend_handles,
              title=label,
              loc='lower left',
              bbox_to_anchor=legend_imp_xy,
              frameon=False,
              title_fontsize=fontsize_title,
              fontsize=fontsize_annotations,
              labelspacing=0.25,
              columnspacing=0, handletextpad=0, handlelength=0,
              borderpad=0)


# II Main Functions
def plot_feature_map(df_feat=None, df_cat=None,
                     col_cat="subcategory", col_val="mean_dif", col_imp="feat_importance",
                     name_test="TEST", name_ref="REF",
                     figsize=(8, 8),
                     start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                     tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                     tmd_color="mediumspringgreen", jmd_color="blue",
                     tmd_seq_color="black", jmd_seq_color="white",
                     seq_size=None,
                     fontsize_tmd_jmd=None, weight_tmd_jmd="normal",
                     fontsize_titles=11, fontsize_labels=12,
                     fontsize_annotations=11, fontsize_imp_bar=9,
                     add_xticks_pos=False,
                     grid_linewidth=0.01, grid_linecolor=None,
                     border_linewidth=2,
                     facecolor_dark=False, vmin=None, vmax=None,
                     cmap=None, cmap_n_colors=101,
                     cbar_pct=True, cbar_kws=None, cbar_xywh=(0.5, None, 0.2, None),
                     dict_color=None, legend_kws=None, legend_xy=(-0.1, -0.01),
                     legend_imp_xy=(1.25, 0),
                     imp_ths=(0.2, 0.5, 1), imp_marker_sizes=(3, 5, 8),
                     imp_bar_th=None,
                     xtick_size=11.0, xtick_width=2.0, xtick_length=5.0):
    """Create a comprehensive feature map with a heatmap, feature importance bars, and custom legends."""
    # Get fontsize
    pe = PlotElements()
    fs = ut.plot_gco()
    fs_titles = fs-1 if fontsize_titles is None else fontsize_titles
    fs_labels = fs if fontsize_labels is None else fontsize_labels
    fs_annotations = fs-1 if fontsize_annotations is None else fontsize_annotations
    fs_imp_bar = fs-3 if fontsize_imp_bar is None else fontsize_imp_bar

    # Group arguments
    args_seq = dict(jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq)
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_part_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    args_seq_color = dict(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
    args_fs = dict(seq_size=seq_size,
                   fontsize_labels=fs_labels,
                   fontsize_tmd_jmd=fontsize_tmd_jmd)
    args_xtick = dict(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)

    # Plot
    width, height = figsize
    width_ratio = [6, min(1*height/width, 1)]
    fig, axes = plt.subplots(figsize=figsize,
                             nrows=1, ncols=2,
                             sharey=True,
                             gridspec_kw={'width_ratios': width_ratio, "wspace": 0},
                             layout="constrained")

    # Set colorbar and legend arguments
    label_cbar = f"Feature value\n{name_test} - {name_ref}"
    _cbar_kws, cbar_ax = pe.adjust_cbar_kws(fig=fig,
                                            cbar_kws=cbar_kws,
                                            cbar_xywh=cbar_xywh,
                                            label=label_cbar,
                                            fontsize_labels=fs_labels)

    n_cat = len(set(df_feat[ut.COL_CAT]))
    _legend_kws = pe.adjust_cat_legend_kws(legend_kws=legend_kws,
                                           n_cat=n_cat,
                                           legend_xy=legend_xy,
                                           fontsize_labels=fs_labels)
    # Plot feat importance bars
    label_bars_imp = "Cumulative feature\nimportance [%]"
    plot_feat_importance_bars(ax=axes[1], df_feat=df_feat.copy(),
                              col_imp=col_imp, col_cat=col_cat,
                              label=label_bars_imp,
                              annotation_th=imp_bar_th,
                              fontsize_label=fs_titles,
                              fontsize_imp_bar=fs_imp_bar,
                              fontsize_annotations=fontsize_annotations,
                              weight_annotation="bold")
    # Plot heatmap
    ax = plot_heatmap_(df_feat=df_feat, df_cat=df_cat,
                       col_cat=col_cat, col_val=col_val,
                       ax=axes[0], figsize=figsize,
                       start=start, **args_len, **args_seq,
                       **args_part_color, **args_seq_color,
                       **args_fs, weight_tmd_jmd=weight_tmd_jmd,
                       add_xticks_pos=add_xticks_pos,
                       grid_linewidth=grid_linewidth, grid_linecolor=grid_linecolor,
                       border_linewidth=border_linewidth,
                       facecolor_dark=facecolor_dark, vmin=vmin, vmax=vmax,
                       cmap=cmap, cmap_n_colors=cmap_n_colors,
                       cbar_ax=cbar_ax, cbar_pct=cbar_pct, cbar_kws=_cbar_kws,
                       dict_color=dict_color, legend_kws=_legend_kws,
                       **args_xtick)

    # Add feature position title
    plt.title(ut.LABEL_FEAT_POS, x=0, weight="bold", fontsize=fs_titles)


    # Add feature importance map
    add_feat_importance_map(df_feat=df_feat, df_cat=df_cat,
                            col_cat=col_cat, col_imp=col_imp,
                            imp_ths=imp_ths,
                            imp_marker_size=imp_marker_sizes,
                            ax=ax, start=start, args_len=args_len)

    legend_imp_xy_default = (1.25, 0)
    _legend_imp_xy = ut.adjust_tuple_elements(tuple_in=legend_imp_xy,
                                              tuple_default=legend_imp_xy_default)
    label_feat_imp = "Feature importance"
    add_feat_importance_legend(ax=cbar_ax,
                               imp_ths=imp_ths,
                               legend_imp_xy=_legend_imp_xy,
                               label=label_feat_imp,
                               fontsize_title=fs_labels,
                               fontsize_annotations=fs_annotations)


    return fig, ax



"""
This is a script for the common AAclust checking functions
"""
import aaanalysis.utils as ut


# I Helper Functions


# II Main Functions
def check_metric(metric=None):
    """"""
    if metric not in ut.LIST_METRICS:
        error = f"'metric' should be None or one of following: {ut.LIST_METRICS}"
        raise ValueError(error)





"""
This is a script for the common SequenceFeature and CPP checking functions
"""
import pandas as pd
import numpy as np
import warnings

import aaanalysis.utils as ut
from .cpp.utils_feature import get_part_positions


# Helper functions
def _get_max_pos_split(split=None):
    """Get maximum position requiered for split"""
    if ut.STR_SEGMENT in split:
        n_max = int(split.split(",")[1].replace(")", ""))
    elif ut.STR_PERIODIC_PATTERN in split:
        n_max = int(split.split(",")[-1].replace(")", ""))
    elif ut.STR_PATTERN:
        n_max = int(split.split(",")[-1].replace(")", ""))
    else:
        raise ValueError(f"Wrong 'split' ({split})")
    return n_max


def _get_max_pos_split_kws(split_kws=None):
    """Get maximum position requiered for splits basd on split_kws"""
    list_n_max = []
    if ut.STR_SEGMENT in split_kws:
        n_max = split_kws[ut.STR_SEGMENT]["n_split_max"]
        list_n_max.append(n_max)
    if ut.STR_PATTERN in split_kws:
        n_max = split_kws[ut.STR_PATTERN]["len_max"]
        list_n_max.append(n_max)
    if ut.STR_PERIODIC_PATTERN in split_kws:
        n_max = split_kws[ut.STR_PERIODIC_PATTERN]["steps"][0]
        list_n_max.append(n_max)
    if len(list_n_max) == 0:
        raise ValueError(f"Wrong 'split_kws' ({split_kws})")
    n_max = max(list_n_max)
    return n_max


# II Main Functions
# Check splits
def check_split_kws(split_kws=None, accept_none=True):
    """Check if argument dictionary for splits is a valid input"""
    # Define the expected structure and types for split_kws
    split_kws_types = {ut.STR_SEGMENT: dict(n_split_min=int, n_split_max=int),
                       ut.STR_PATTERN: dict(steps=list, n_min=int, n_max=int, len_max=int),
                       ut.STR_PERIODIC_PATTERN: dict(steps=list)}
    if accept_none and split_kws is None:
        return None     # Skip check
    if not isinstance(split_kws, dict):
        raise ValueError(f"'split_kws' should be type dict (not {split_kws})")
    if len(split_kws) == 0:
        raise ValueError("'split_kws' should be not empty")
    # Check if split_kws contains wrong split_types
    invalid_split_types = [x for x in split_kws if x not in ut.LIST_SPLIT_TYPES]
    if len(invalid_split_types) > 0:
        error = f"Following 'split_types' are invalid: {invalid_split_types}." \
                f"\n  'split_kws' should have following structure: {split_kws_types}."
        raise ValueError(error)
    # Validate split types and argument structure
    for split_type in split_kws:
        expected_args = split_kws_types[split_type]
        missing_args = [arg for arg in expected_args if arg not in split_kws[split_type]]
        if missing_args:
            raise ValueError(f"Missing required arguments for '{split_type}': {missing_args}")
        for arg in split_kws[split_type]:
            if arg not in split_kws_types[split_type]:
                error = f"'{arg}' arg in '{split_type}' of 'split_kws' is invalid." \
                        f"\n  'split_kws' should have following structure: {split_kws_types}."
                raise ValueError(error)
            arg_val = split_kws[split_type][arg]
            expected_arg_type = split_kws_types[split_type][arg]
            if not isinstance(arg_val, expected_arg_type):
                error = f"Type of '{arg}':'{arg_val}' ('{type(arg_val)}') should be '{expected_arg_type}'"
                raise ValueError(error)
            if isinstance(arg_val, list):
                wrong_type = [x for x in arg_val if type(x) is not int]
                if len(wrong_type) > 0:
                    error = f"All list elements ({arg_val}) of '{arg}' should have type int."
                    raise ValueError(error)
    # Check Segment
    if ut.STR_SEGMENT in split_kws:
        segment_args = split_kws[ut.STR_SEGMENT]
        n_split_min, n_split_max = segment_args["n_split_min"], segment_args["n_split_max"]
        ut.check_number_range(name=f"{ut.STR_SEGMENT}[n_split_min]", val=n_split_min, just_int=True, min_val=1)
        ut.check_number_range(name=f"{ut.STR_SEGMENT}[n_split_max]", val=n_split_max, just_int=True, min_val=1)
        if n_split_min > n_split_max:
            raise ValueError(f"For '{ut.STR_SEGMENT}', 'n_split_min' ({n_split_min}) should be smaller "
                             f"or equal to 'n_split_max' ({n_split_max})")
    # Check Pattern
    if ut.STR_PATTERN in split_kws:
        pattern_args = split_kws[ut.STR_PATTERN]
        n_min, n_max = pattern_args["n_min"], pattern_args["n_max"]
        steps_pattern, len_max = pattern_args["steps"], pattern_args["len_max"]
        ut.check_number_range(name=f"{ut.STR_PATTERN}[n_min]", val=n_min, just_int=True, min_val=1)
        ut.check_number_range(name=f"{ut.STR_PATTERN}[n_max]", val=n_max, just_int=True, min_val=1)
        ut.check_number_range(name=f"{ut.STR_PATTERN}[len_max]", val=n_max, just_int=True, min_val=1)
        if n_min > n_max:
            raise ValueError(f"For '{ut.STR_PATTERN}', 'n_min' ({n_min}) should be smaller or equal to 'n_max' ({n_max})")
        if not isinstance(steps_pattern, list) or len(steps_pattern) < 1:
            raise ValueError(f"'steps_pattern' ({steps_pattern}) should be non-empty list of with at"
                             f" least 1 non-negative integers")
        if steps_pattern != sorted(steps_pattern):
            raise ValueError(f"For '{ut.STR_PATTERN}', 'steps_pattern' ({steps_pattern}) should be ordered in ascending order.")
        if steps_pattern[0] >= len_max:
            raise ValueError(f"For '{ut.STR_PATTERN}', 'len_max' ({len_max}) should be greater than the smallest step "
                             f"in 'steps_pattern' ({steps_pattern}).")
        for i, step in enumerate(steps_pattern):
            ut.check_number_range(name=f"{ut.STR_PATTERN}[steps_pattern](step{i+1})", val=step, just_int=True,
                                  min_val=1)
    # Check PeriodicPattern
    if ut.STR_PERIODIC_PATTERN in split_kws:
        periodicpattern_args = split_kws[ut.STR_PERIODIC_PATTERN]
        steps_periodicpattern = periodicpattern_args["steps"]
        if not isinstance(steps_periodicpattern, list) or len(steps_periodicpattern) != 2:
            raise ValueError(f"'steps_periodicpattern' ({steps_periodicpattern}) should be list of"
                             f" with exactly 2 non-negative integers")
        if steps_periodicpattern != sorted(steps_periodicpattern):
            raise ValueError(f"For '{ut.STR_PERIODIC_PATTERN}', 'steps_periodicpattern' ({steps_periodicpattern}) "
                             f"should be ordered in ascending order.")
        step1, step2 = steps_periodicpattern
        ut.check_number_range(name=f"{ut.STR_PERIODIC_PATTERN}[steps_periodicpattern](step1)",
                              val=step1, just_int=True, min_val=1)
        ut.check_number_range(name=f"{ut.STR_PERIODIC_PATTERN}[steps_periodicpattern](step2)",
                              val=step2, just_int=True, min_val=1)


# Check parts
def check_parts_len(tmd_len=None, jmd_n_len=None, jmd_c_len=None,
                    accept_none_tmd_len=False, accept_none_jmd_len=False,
                    tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                    check_jmd_seq_len_consistent=False):
    """Check length parameters and if they are matching with sequences if provided"""
    tmd_seq = ut.check_str(name="tmd_seq", val=tmd_seq, accept_none=True, return_empty_string=True)
    jmd_n_seq = ut.check_str(name="jmd_n_seq", val=jmd_n_seq, accept_none=True, return_empty_string=True)
    jmd_c_seq = ut.check_str(name="jmd_c_seq", val=jmd_c_seq, accept_none=True, return_empty_string=True)
    # If sequences is not None, set length to sequence length
    if len(jmd_n_seq + tmd_seq + jmd_c_seq) > 0:
        if check_jmd_seq_len_consistent:
            if jmd_n_len is None or jmd_n_len != len(jmd_n_seq):
                raise ValueError(f"Not matching of 'jmd_n_len' ({jmd_n_len}) and 'jmd_n_seq' ({jmd_n_seq})")
            if jmd_c_len is None or jmd_c_len != len(jmd_c_seq):
                raise ValueError(f"Not matching of 'jmd_c_len' ({jmd_c_len}) and 'jmd_n_seq' ({jmd_c_seq})")
        tmd_len, jmd_n_len, jmd_c_len = len(tmd_seq), len(jmd_n_seq), len(jmd_c_seq)
    else:
        tmd_seq = jmd_n_seq = jmd_c_seq = None
    # Check lengths
    ext_len = ut.options["ext_len"]
    ut.check_number_range(name="tmd_len", val=tmd_len, accept_none=accept_none_tmd_len, min_val=1, just_int=True)
    ut.check_number_range(name="jmd_n_len", val=jmd_n_len, accept_none=accept_none_jmd_len, min_val=0, just_int=True)
    ut.check_number_range(name="jmd_c_len", val=jmd_c_len, accept_none=accept_none_jmd_len, min_val=0, just_int=True)
    ut.check_number_range(name="ext_len", val=ext_len, min_val=0, accept_none=True, just_int=True)
    # Check len_ext
    if ext_len is not None and ext_len != 0:
        # Check if ext_len exceeds either jmd_n_len or jmd_c_len
        if ext_len > jmd_n_len:
            raise ValueError(f"'ext_len' ({ext_len}) must be <= length of jmd_n ({jmd_n_len})")
        if ext_len > jmd_c_len:
            raise ValueError(f"'ext_len' ({ext_len}) must be <= length of jmd_c ({jmd_c_len})")
    # Create out.dictionaries
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_seq = dict(tmd_seq=tmd_seq, jmd_n_seq=jmd_n_seq, jmd_c_seq=jmd_c_seq)
    return args_len, args_seq


def check_match_features_seq_parts(features=None, tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                                   tmd_len=20, jmd_n_len=10, jmd_c_len=10):
    """Check if sequence lengths do match with length requirements of features"""
    # Check match of part length and features
    if None in [tmd_seq, jmd_n_seq, jmd_c_seq] or len(jmd_n_seq + tmd_seq + jmd_c_seq) == 0:
        jmd_n, tmd, jmd_c = get_part_positions(start=1, jmd_n_len=jmd_n_len, tmd_len=tmd_len, jmd_c_len=jmd_c_len)
        dict_part_seq = ut.get_dict_part_seq(tmd=tmd, jmd_n=jmd_n, jmd_c=jmd_c)
        for feature in features:
            part, split, scale = feature.split("-")
            n_max = _get_max_pos_split(split=split)
            seq = dict_part_seq[part.lower()]
            if len(seq) < n_max:
                raise ValueError(f"Sequence length (n={len(seq)}) too short for '{feature}' feature (n_max={n_max})")
    # Check match of sequence part length and features
    else:
        dict_part_seq = ut.get_dict_part_seq(tmd=tmd_seq, jmd_c=jmd_c_seq, jmd_n=jmd_n_seq)
        for feature in features:
            part, split, scale = feature.split("-")
            n_max = _get_max_pos_split(split=split)
            seq = dict_part_seq[part.lower()]
            if len(seq) < n_max:
                raise ValueError(
                    f"Sequence 'part' ({seq}, n={len(seq)}) too short for '{feature}' feature (n_max={n_max})")


# Check df_seq
def _get_tmd_positions(row):
    """Get position of tmd from sequence"""
    tmd, seq = row[ut.COL_TMD], row[ut.COL_SEQ]
    tmd_start = seq.find(tmd)
    tmd_stop = tmd_start + len(tmd) if tmd_start != -1 else -1
    if tmd_start == -1 or tmd_start == tmd_stop:
        raise ValueError(f"'{ut.COL_TMD}' is not contained in '{ut.COL_SEQ}' for '{row[ut.COL_ENTRY]}' entry")
    tmd_start += 1
    return pd.Series([tmd_start, tmd_stop])


def check_match_df_seq_jmd_len(df_seq=None, jmd_n_len=None, jmd_c_len=None):
    """Check matching of df_seq and jmd lengths."""
    df_seq = df_seq.copy()
    pos_based = set(ut.COLS_SEQ_POS).issubset(set(df_seq))
    part_based = set(ut.COLS_SEQ_PARTS).issubset(set(df_seq))
    seq_based = ut.COL_SEQ in list(df_seq)
    seq_tmd_based = set(ut.COLS_SEQ_TMD).issubset(set(df_seq))
    # 'jmd_n_len' and 'jmd_c_len' should be already checked if None or int by interface
    if [jmd_n_len, jmd_c_len].count(None) >= 1:
        if not part_based:
            raise ValueError(f"'jmd_n_len' and 'jmd_c_len' should be both given if '{ut.COLS_SEQ_PARTS}' are not provided")
        else:
            if [jmd_n_len, jmd_c_len].count(None) == 1:
                raise ValueError(f"'jmd_n_len' and 'jmd_c_len' should be both given or both None "
                                 f"if '{ut.COLS_SEQ_PARTS}' are provided")
    # Get 'tmd_start' and 'tmd_stop'
    elif [jmd_n_len, jmd_c_len].count(None) == 0:
        if part_based and not pos_based:
            df_seq[ut.COL_SEQ] = df_seq[ut.COL_JMD_N] + df_seq[ut.COL_TMD] + df_seq[ut.COL_JMD_C]
            df_seq[[ut.COL_TMD_START, ut.COL_TMD_STOP]] = df_seq.apply(_get_tmd_positions, axis=1)
    if not pos_based and not part_based:
        if seq_tmd_based:
            df_seq[[ut.COL_TMD_START, ut.COL_TMD_STOP]] = df_seq.apply(_get_tmd_positions, axis=1)
        elif seq_based:
            tmd_start = 1 + jmd_n_len
            list_seq = []
            list_tmd_stop = []
            for seq in df_seq[ut.COL_SEQ]:
                # If 'jmd_n_len' and 'jmd_c_len' exceed the sequence length, sequence is adjusted using gaps.
                dif_jmd_n_len_seq = jmd_n_len - len(seq)
                if dif_jmd_n_len_seq >= 0:
                    # Add one gap for TMD
                    seq += ut.STR_AA_GAP * (dif_jmd_n_len_seq + 1)
                dif_jmd_len_seq = jmd_c_len + jmd_n_len - len(seq)
                if dif_jmd_len_seq >= 0:
                    # If no jmd_n, add gaps to N-terminus of sequence
                    if jmd_n_len == 0:
                        seq = ut.STR_AA_GAP * (dif_jmd_len_seq + 1) + seq
                    # If jmd_n, add gaps to C-terminus of sequence
                    else:
                        seq += ut.STR_AA_GAP * (dif_jmd_len_seq + 1)
                    tmd_stop = tmd_start
                else:
                    tmd_stop = len(seq) - jmd_c_len
                list_seq.append(seq)
                list_tmd_stop.append(tmd_stop)
            df_seq[ut.COL_TMD_START] = tmd_start
            df_seq[ut.COL_TMD_STOP] = list_tmd_stop
            df_seq[ut.COL_SEQ] = list_seq
    return df_seq


# Check df_parts
def check_df_parts(df_parts=None, accept_none=False):
    """Check if df_parts is a valid input"""
    ut.check_df(name="df_parts", df=df_parts, accept_none=accept_none)
    if df_parts is None and accept_none:
        return # Skip check
    if len(list(df_parts)) == 0 or len(df_parts) == 0:
        raise ValueError("'df_parts' should not be empty pd.DataFrame")
    ut.check_list_parts(list_parts=list(df_parts))
    # Check if columns are unique
    if len(list(df_parts)) != len(set(df_parts)):
        raise ValueError("Column names in 'df_parts' must be unique. Drop duplicates!")
    # Check if index is unique
    if len(list(df_parts.index)) != len(set(df_parts.index)):
        raise ValueError("Index in 'df_parts' must be unique. Drop duplicates!")
    # Check if columns contain strings
    dict_dtype = dict(df_parts.dtypes)
    cols_wrong_type = [col for col in dict_dtype if dict_dtype[col] not in [object, str]]
    if len(cols_wrong_type) > 0:
        error = "'df_parts' should contain sequences with type string." \
                f"\n  Following columns contain no values with type string: {cols_wrong_type}"
        raise ValueError(error)


def check_match_df_parts_features(df_parts=None, features=None):
    """Check if df_parts does match with length requirements of features"""
    if df_parts is None:
        return # Skip check
    for feature in features:
        part, split, scale = feature.split("-")
        n_min = _get_max_pos_split(split=split)
        if any(df_parts[part.lower()].map(len) < n_min):
            mask = df_parts[part.lower()].map(len) < n_min
            list_seq = df_parts[mask][part.lower()].to_list()
            if len(list_seq) == 1:
                seq = list_seq[0]
                raise ValueError(f"Sequence 'part' ('{seq}', n={len(seq)}) is too short"
                                 f"\n  for '{feature}' feature (n_min={n_min})")
            else:
                raise ValueError(
                    f"For '{feature}' feature (n_min={n_min}),"
                    f"\n  following sequence 'parts' are too short: {list_seq}")


def check_match_df_parts_list_parts(df_parts=None, list_parts=None):
    """"""
    if list_parts is not None and df_parts is not None:
        list_parts = ut.check_list_like(name="list_parts", val=list_parts, accept_str=True, convert=True)
        missing_parts = [part.lower() for part in list_parts if part.lower() is not list(df_parts)]
        raise ValueError(f"'part' ({missing_parts}) must be in columns of 'df_parts': {list(df_parts)}")


def check_match_df_parts_split_kws(df_parts=None, split_kws=None):
    """Check if df_parts and split_kws match regarding the sequence size"""
    n_max = _get_max_pos_split_kws(split_kws=split_kws)
    for part in list(df_parts):
        if any(df_parts[part.lower()].map(len) < n_max):
            mask = df_parts[part.lower()].map(len) < n_max
            list_seq = df_parts[mask][part.lower()].to_list()
            if len(list_seq) == 1:
                seq = list_seq[0]
                raise ValueError(
                    f"'{part}' part contains too short sequence ('{seq}', n={len(seq)})"
                    f"\n  for '{split_kws}' split_kws (n_max={n_max})")
            else:
                seq = list_seq[0]
                raise ValueError(
                    f"For split_kws (n_max={n_max}): '{split_kws}',"
                    f"\n  following '{part}' part contains too short sequences (e.g., '{seq}', n={len(seq)}).")


# Check df_scales & df_cat
def check_df_scales(df_scales=None, accept_none=False):
    """Check if df_scales is a valid input and matching to df_parts"""
    ut.check_df(name="df_scales", df=df_scales, accept_none=accept_none)
    if df_scales is None and accept_none:
        return  # Skip check
    # Check if columns are unique
    if len(list(df_scales)) != len(set(df_scales)):
        raise ValueError("Column names in 'df_scales' must be unique. Drop duplicates!")
    # Check if index is unique
    if len(list(df_scales.index)) != len(set(df_scales.index)):
        raise ValueError("Index in 'df_scales' must be unique. Drop duplicates!")
    # Check if columns contain numbers
    dict_dtype = dict(df_scales.dtypes)
    cols_wrong_type = [col for col in dict_dtype if not np.issubdtype(dict_dtype[col], np.number)]
    if len(cols_wrong_type) > 0:
        error = "'df_scales' should only contain numbers." \
                f"\n  Following columns contain no numerical values: {cols_wrong_type}"
        raise ValueError(error)
    # Check if NaN in df
    cols_nans = [x for x in list(df_scales) if df_scales[x].isnull().any()]
    if len(cols_nans) > 0:
        error = "'df_scales' should not contain NaN." \
                f"\n  Following columns contain NaN: {cols_nans}"
        raise ValueError(error)


def check_match_df_scales_features(df_scales=None, features=None):
    """Check if scale ids from df_scales does match with features"""
    if df_scales is None:
        return # Skip check
    scales_feat = [x.split("-")[2] for x in features]
    scales = list(df_scales)
    missing_scales_in_df_scales = [x for x in scales_feat if x not in scales]
    if len(missing_scales_in_df_scales) > 0:
        raise ValueError(f"Following scale ids are missing in 'df_scales': {missing_scales_in_df_scales}")


# Check df_cat
def check_df_cat(df_cat=None, accept_none=True):
    """Check if df_cat is valid"""
    if df_cat is None and accept_none:
        return # Skip check
    cols_cat = [ut.COL_SCALE_ID, ut.COL_CAT, ut.COL_SUBCAT]
    ut.check_df(name="df_cat", df=df_cat, cols_requiered=cols_cat, accept_none=accept_none)


def check_match_df_cat_features(df_cat=None, features=None):
    """Check if scale ids from df_cat does match with features"""
    if df_cat is None:
        return # Skip check
    scales_feat = [x.split("-")[2] for x in features]
    scales_cat = list(df_cat[ut.COL_SCALE_ID])
    missing_scales_in_df_cat = [x for x in scales_feat if x not in scales_cat]
    if len(missing_scales_in_df_cat) > 0:
        raise ValueError(f"Following scale ids are missing in 'df_cat': {missing_scales_in_df_cat}")


# Check matching of df_scales with df_parts and df_scales
def check_match_df_parts_df_scales(df_parts=None, df_scales=None, accept_gaps=False):
    """Check if characters from df_parts match with scales from df_scales"""
    if df_parts is not None and df_scales is not None:
        f = lambda x: set(x)
        vf = np.vectorize(f)
        char_parts = set().union(*vf(df_parts.values).flatten())
        char_scales = list(set(df_scales.index))
        if accept_gaps:
            char_scales.append(ut.STR_AA_GAP)
        missing_char = [x for x in char_parts if x not in char_scales]
        # Replace gaps by default amino acid gap
        if accept_gaps:
            for col in list(df_parts):
                for mc in missing_char:
                    df_parts[col] = df_parts[col].str.replace(mc, ut.STR_AA_GAP)
        elif len(missing_char) > 0:
            error = f"Not all characters in sequences from 'df_parts' are covered!"\
                    f"\n  Following characters are missing in 'df_scales': {missing_char}." \
                    f"\n  Consider enabling 'accept_gaps'"
            raise ValueError(error)
    return df_parts


def check_match_df_scales_df_cat(df_cat=None, df_scales=None, verbose=True):
    """Check if scale ids matches from df_cat and df_scales"""
    if df_scales is not None and df_cat is not None:
        scales_cat = list(df_cat[ut.COL_SCALE_ID])
        scales = list(df_scales)
        overlap_scales = [x for x in scales if x in scales_cat]
        difference_scales = list(set(scales).difference(set(scales_cat)))
        # Adjust df_cat and df_scales
        df_cat = df_cat[df_cat[ut.COL_SCALE_ID].isin(overlap_scales)]
        df_scales = df_scales[overlap_scales]
        if len(difference_scales) > 0:
            missing_scales_in_df_scales = [x for x in scales_cat if x not in scales]
            missing_scales_in_df_cat = [x for x in scales if x not in scales_cat]
            # Error since all scale ids must be covered by df_cat
            if missing_scales_in_df_cat:
                raise ValueError(f"Following scale ids from 'df_scales' are missing in 'df_cat': {missing_scales_in_df_cat}")
            # Only warning (excepted)
            if verbose and len(missing_scales_in_df_scales) > 0:
                warnings.warn(f"Following scale ids from 'df_cat' are missing in 'df_scales': {missing_scales_in_df_scales}")
    return df_scales, df_cat



"""
This is a script for the backend of the CPPPlot.profile() method.
"""
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import warnings

import aaanalysis.utils as ut
from ._utils_cpp_plot_elements import PlotElements
from ._utils_cpp_plot_positions import PlotPartPositions


# I Helper Functions
def _scale_ylim(df_bars=None, ylim=None, col_imp=None, retrieve_plot=False, scaling_factor=1.1, verbose=True):
    """Adjust y-axis limits based on data range and provided ylim."""
    if ylim is not None:
        max_val = round(max(df_bars[col_imp]), 3)
        min_val = round(min(df_bars[col_imp]), 3)
        min_y, max_y = ylim
        warning_msg = ""
        if min_val < min_y and max_val > max_y:
            warning_msg = f"The provided 'ylim' ({ylim}) is both smaller and larger than the data range ({min_val}, {max_val})."
            ylim = (min_val, max_val * scaling_factor)
        elif min_val < min_y:
            warning_msg = f"The minimum value of 'ylim' ({min_y}) is larger than the minimum data value ({min_val})."
            ylim = (min_val, max_y)
        elif max_val > max_y:
            warning_msg = f"The maximum value of 'ylim' ({max_y}) is smaller than the maximum data value ({max_val})."
            ylim = (min_y, max_val * scaling_factor)
        if warning_msg and verbose:
            warnings.warn(warning_msg, UserWarning)
    else:
        if retrieve_plot:
            ylim = plt.ylim()
            ylim = (ylim[0] * scaling_factor, ylim[1] * scaling_factor)
    return ylim


# II Main Functions
# CPP/-SHAP profile plotting functions
def _plot_cpp_profile(ax=None, df_pos=None, dict_color=None, show_cat=True, color=None, ylim=None,
                      plot_args=None):
    """Plot CPP profile on given axes."""
    labels = list(dict_color.keys())
    df_bar = df_pos.T[labels]
    df = pd.concat([df_bar[df_bar < 0].sum(axis=1), df_bar[df_bar > 0].sum(axis=1)]).to_frame(name="col_cat")
    color = dict_color if show_cat else color
    if not show_cat:
        df_bar = df_bar.sum(axis=1)
    ax = df_bar.plot(ax=ax, color=color, **plot_args)
    ylim = _scale_ylim(df_bars=df, col_imp="col_cat", ylim=ylim)
    plt.ylim(ylim)
    return ax


def _plot_cpp_shap_profile(ax=None, df_pos=None, ylim=None, plot_args=None):
    """Plot CPP-SHAP profile on given axes."""
    df_bar = df_pos.T
    df = pd.concat([df_bar[df_bar < 0].sum(axis=1), df_bar[df_bar > 0].sum(axis=1)]).to_frame(name="col_cat")
    df_pos = df_bar[df_bar > 0]
    df_pos = df_pos.sum(axis=1)
    df_neg = df_bar[df_bar < 0]
    df_neg = df_neg.sum(axis=1)
    ax = df_pos.plot(ax=ax, color=ut.COLOR_SHAP_POS, **plot_args)
    ax = df_neg.plot(ax=ax, color=ut.COLOR_SHAP_NEG, **plot_args)
    ylim = _scale_ylim(df_bars=df, col_imp="col_cat", ylim=ylim, retrieve_plot=True)
    plt.ylim(ylim)
    return ax


# Inner plotting function
def _plot_profile(df_pos=None, shap_plot=False, ax=None,
                  start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                  show_cat=True, dict_color=None,
                  bar_width=0.8, edge_color="none",
                  color="tab:gray", ylim=None,
                  xtick_size=11.0, xtick_width=2.0, xtick_length=5.0,
                  ytick_size=None, ytick_width=None, ytick_length=5.0):
    """Inner function to plot feature profile by calling CPP or CPP-SHAP, and set ticks and limits"""
    # Constants for additional spacing
    XLIM_ADD = 3 if shap_plot else 1
    seq_len = jmd_n_len + tmd_len + jmd_c_len
    pp = PlotPartPositions(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, start=start)
    plot_args = dict(kind="bar", stacked=True, rot=0, width=bar_width, edgecolor=edge_color,
                     legend=False, zorder=10)

    # Plot based on SHAP values or not
    if shap_plot:
        ax = _plot_cpp_shap_profile(ax=ax, df_pos=df_pos, plot_args=plot_args, ylim=ylim)
    else:
        ax = _plot_cpp_profile(ax=ax, df_pos=df_pos, plot_args=plot_args,
                               ylim=ylim, dict_color=dict_color,
                               show_cat=show_cat, color=color)

    # Setting x-ticks and y-ticks
    xticks, xticks_labels = pp.get_xticks_with_labels(step=5)
    ax.tick_params(axis="x", color="black", width=xtick_width, length=xtick_length)
    ax.set_xticks(xticks)
    if xtick_size > 0:
        ax.set_xticklabels(xticks_labels, rotation=0, size=xtick_size)
    plt.yticks(size=ytick_size)
    ax.tick_params(axis="y", color="black", width=ytick_width, length=ytick_length, bottom=False)

    # Add extra flanking space for axes
    x_lim = (min(xticks) - XLIM_ADD, max(xticks) + XLIM_ADD)
    ax.set_xlim(x_lim)
    if ylim is None:
        ymin, ymax = ax.get_ylim()
        y_space = min(0, (ymax - ymin) * 0.25)
        y_lim = (ymin - y_space, ymax + y_space)
        ax.set_ylim(y_lim)

    # Plot baseline for reference
    ax.plot([-0.5, seq_len - 0.5], [0, 0], color="black", linestyle="-")
    sns.despine(top=True, right=True)
    return ax


# Main plotting function
def plot_profile(df_feat=None, df_cat=None, shap_plot=False,
                 col_imp="feat_importance", normalize=False,
                 ax=None, figsize=(7, 5),
                 start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                 jmd_n_seq=None, tmd_seq=None, jmd_c_seq=None,
                 tmd_color="mediumspringgreen", jmd_color="blue",
                 tmd_seq_color="black", jmd_seq_color="white",
                 seq_size=None, fontsize_tmd_jmd=None, weight_tmd_jmd="normal",
                 add_xticks_pos=False, highlight_tmd_area=True, highlight_alpha=0.15,
                 add_legend_cat=True, dict_color=None,legend_kws=None,
                 bar_width=0.75, edge_color="none",
                 grid_axis=None, ylim=None,
                 xtick_size=11.0, xtick_width=2.0, xtick_length=5.0,
                 ytick_size=None, ytick_width=None, ytick_length=5.0):
    """Main function to plot feature profiles with various customizations and styling."""
    # Group arguments
    args_seq = dict(jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq)
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_part_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    args_seq_color = dict(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
    args_xtick = dict(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)
    args_ytick= dict(ytick_size=ytick_size, ytick_width=ytick_width, ytick_length=ytick_length)
    # Get df positions
    value_type = "sum" if col_imp else "count"
    col_cat = "scale_name" if shap_plot else "category"
    pp = PlotPartPositions(**args_len, start=start)
    df_pos = pp.get_df_pos(df_feat=df_feat.copy(), df_cat=df_cat.copy(),
                           col_cat=col_cat, col_val=col_imp,
                           value_type=value_type, normalize=normalize)
    # Plotting
    pe = PlotElements()
    fig, ax = pe.set_figsize(ax=ax, figsize=figsize, force_set=True)
    ax = _plot_profile(df_pos=df_pos, shap_plot=shap_plot, ax=ax,
                       start=start, **args_len,
                       show_cat=add_legend_cat, dict_color=dict_color,
                       edge_color=edge_color, bar_width=bar_width,
                       ylim=ylim, **args_ytick, **args_xtick)
    # Set default ylabel
    if col_imp is None:
        ylabel = ut.LABEL_FEAT_NUMBER
    else:
        ylabel = ut.LABEL_FEAT_IMPACT_CUM if shap_plot else ut.LABEL_FEAT_IMPORT_CUM
    ax.set_ylabel(ylabel)
    # Add grid
    if grid_axis is not None:
        ax.set_axisbelow(True)  # Grid behind datasets
        ax.grid(which="major", axis=grid_axis, linestyle="-")
    # Add tmd area
    if highlight_tmd_area:
        pp.highlight_tmd_area(ax=ax, x_shift=-0.5, tmd_color=tmd_color, alpha=highlight_alpha)
    # Add TMD-JMD sequence
    if type(tmd_seq) == str:
        pp.add_tmd_jmd_seq(ax=ax, **args_seq, **args_part_color, **args_seq_color,
                           seq_size=seq_size,
                           fontsize_tmd_jmd=fontsize_tmd_jmd, weight_tmd_jmd=weight_tmd_jmd,
                           add_xticks_pos=add_xticks_pos, heatmap=False,
                           x_shift=0, xtick_size=xtick_size)
    # Add TMD-JMD bar
    else:
        pp.add_tmd_jmd_bar(ax=ax, x_shift=-0.5, **args_part_color)
        pp.add_tmd_jmd_xticks(ax=ax, x_shift=0, **args_xtick)
        pp.add_tmd_jmd_text(ax=ax, x_shift=-0.5,
                            fontsize_tmd_jmd=fontsize_tmd_jmd,
                            weight_tmd_jmd=weight_tmd_jmd)
    # Add legend
    if add_legend_cat:
        fs = ut.plot_gco() - 2
        _legend_kws = dict(n_cols=2, loc=2, fontsize=fs,
                           labelspacing=0.05, handletextpad=0.2,
                           fontsize_title=fs, title=ut.LABEL_SCALE_CAT,
                           weight_title="bold")
        if legend_kws is not None:
            _legend_kws.update(legend_kws)
        ut.plot_legend_(ax=ax, dict_color=dict_color, **_legend_kws)
    # Set current axis to main axis object depending on tmd sequence given or not
    plt.sca(plt.gcf().axes[0])
    ax = plt.gca()
    return fig, ax



"""
This is a script for frontend of the plotting utility function to obtain an AAanalysis color dict.
The backend is in general utility module to provide function to remaining AAanalysis modules.
"""
from aaanalysis import utils as ut


# Main function
def plot_get_cdict(name: str = "DICT_COLOR") -> dict:
    """
    Get color dictionaries specified for AAanalysis.

    Parameters
    ----------
    name : {'DICT_COLOR', 'DICT_CAT'}, default='DICT_COLOR'
        The name of the AAanalysis color dictionary.

         - ``DICT_COLOR``: Dictionary with default colors for plots.
         - ``DICT_CAT``: Dictionary with default colors for scale categories.

    Returns
    -------
    dict_color
       AAanalysis color dictionary.

    See Also
    --------
    * `Plotting Prelude <plotting_prelude.html>`_.

    Examples
    --------
    .. include:: examples/plot_get_cdict.rst
    """
    # Check input
    list_names = [ut.STR_DICT_COLOR, ut.STR_DICT_CAT]
    if name not in list_names:
        raise ValueError(f"'name' must be one of following: {list_names}")
    # Get plotting dictionary
    dict_color = ut.plot_get_cdict_(name=name)
    return dict_color



"""
This is a script for utility functions for statistical measures.
"""
import numpy as np
from sklearn.metrics import roc_auc_score
from scipy.stats import entropy, gaussian_kde
from collections import OrderedDict
from scipy.spatial import distance


# AUC adjusted
def auc_adjusted_(X=None, labels=None, label_test=1):
    """Get adjusted Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    comparing, for each feature, groups (given by y (labels)) by feature values in X (feature matrix).
    """
    # Convert labels to binary format (1 or 0)
    labels_binary = [1 if y == label_test else 0 for y in labels]
    auc = np.apply_along_axis((lambda x: roc_auc_score(labels_binary, x) - 0.5), 0, X)
    auc = np.round(auc, 3)
    return auc


# Bayesian Information Criterion for clusters
def _cluster_center(X):
    """Compute cluster center (i.e., arithmetical mean over all data points/observations of a cluster)"""
    return X.mean(axis=0)[np.newaxis, :]


def _compute_centers(X, labels=None):
    """Obtain cluster centers and their labels"""
    labels_centers = list(OrderedDict.fromkeys(labels))
    list_masks = [[i == label for i in labels] for label in labels_centers]
    centers = np.concatenate([_cluster_center(X[mask]) for mask in list_masks]).round(3)
    labels_centers = np.array(labels_centers)
    return centers, labels_centers


def bic_score_(X, labels=None):
    """Computes the Bayesian Information Criterion (BIC) metric for given clusters."""
    epsilon = 1e-10 # prevent division by zero

    # Check if labels match to number of clusters
    n_classes = len(set(labels))
    n_samples, n_features = X.shape
    if n_classes >= n_samples:
        raise ValueError(f"Number of classes in 'labels' ({n_classes}) must be smaller than n_samples ({n_samples})")
    if n_features == 0:
        raise ValueError(f"'n_features' should not be 0")

    # Map labels to increasing order starting with 0
    unique_labels, inverse = np.unique(labels, return_inverse=True)
    labels = inverse
    centers, center_labels = _compute_centers(X, labels=labels)
    size_clusters = np.bincount(labels)

    # Compute variance over all clusters
    list_masks = [labels == label for label in center_labels]
    sum_squared_dist = sum([sum(distance.cdist(X[mask], [center], 'euclidean') ** 2) for mask, center in zip(list_masks, centers)])

    # Compute between-cluster variance
    denominator = max((n_samples - n_classes) * n_features, epsilon)
    bet_clu_var = max((1.0 / denominator) * sum_squared_dist, epsilon)

    # Compute BIC components
    const_term = 0.5 * n_classes * np.log(n_samples) * (n_features + 1)

    log_size_clusters = np.log(size_clusters + epsilon)
    log_n_samples = np.log(n_samples + epsilon)
    log_bcv = np.log(2 * np.pi * bet_clu_var)

    bic_components = size_clusters * (log_size_clusters - log_n_samples) - 0.5 * size_clusters * n_features * log_bcv - 0.5 * (size_clusters - 1) * n_features
    bic = np.sum(bic_components) - const_term
    return bic


# Kullback-Leibler Divergence
def _comp_kld_for_feature(args):
    """Compute KLD for a single feature."""
    x1, x2 = args
    kde1 = gaussian_kde(x1)
    kde2 = gaussian_kde(x2)
    xmin = min(x1.min(), x2.min())
    xmax = max(x1.max(), x2.max())
    x = np.linspace(xmin, xmax, 1000)
    density1 = kde1(x)
    density2 = kde2(x)
    return entropy(density1, density2)


def kullback_leibler_divergence_(X=None, labels=None, label_test=0, label_ref=1):
    """Calculate the average Kullback-Leibler Divergence (KLD) for each feature."""
    mask_test = np.asarray([x == label_test for x in labels])
    mask_ref = np.asarray([x == label_ref for x in labels])
    X1 = X[mask_ref]
    X2 = X[mask_test]
    # Prepare arguments for each feature
    args = [(X1[:, i], X2[:, i]) for i in range(X.shape[1])]
    # Compute KLD for each feature
    kld = np.array([_comp_kld_for_feature(arg) for arg in args])
    return kld



"""
Script for (backend class) Part object used to retrieve sequence parts for given sequences.
"""
import aaanalysis.utils as ut


# II Main Functions
class Parts:
    """
    Class for creating all sequence parts.

    Parameters
    ----------
    seq: sequence (e.g., amino acids sequence for protein)
    tmd_start: start position within the sequence of TMD
    tmd_stop: stop position within the sequence of the TMD
    jmd_n_len: length of JMD-N
    jmd_c_len: length of JMD-C
    """
    def __init__(self, seq=None, tmd_start=None, tmd_stop=None, jmd_n_len=10, jmd_c_len=10):
        self._seq = seq
        self._tmd_start = int(tmd_start)
        self._tmd_stop = int(tmd_stop)
        self._jmd_n_len = jmd_n_len
        self._jmd_c_len = jmd_c_len
        self._n_terminus_len = self._tmd_start - 1
        self._c_terminus_len = len(self._seq) - self._tmd_stop
        self.tmd = self._tmd()
        self.jmd_n = self._jmd_n()
        self.jmd_c = self._jmd_c()

    def _tmd(self):
        """Get sequence of TMD"""
        start = self._tmd_start - 1
        stop = self._tmd_stop
        return self._seq[start:stop]

    def _jmd_n(self):
        """Get sequence of N-terminal JMD"""
        if self._n_terminus_len >= self._jmd_n_len:
            start = self._tmd_start - (self._jmd_n_len + 1)
            stop = self._tmd_start - 1
            jmd_n = self._seq[start:stop]
        else:
            start = 0
            stop = self._tmd_start - 1
            part = self._seq[start:stop]
            jmd_n = ut.STR_AA_GAP * (self._jmd_n_len - len(part)) + part  # Add "-" (gap) for missing AA in JMD
        return jmd_n

    def _jmd_c(self):
        """Get sequence of C-terminal JMD"""
        if self._c_terminus_len >= self._jmd_c_len:
            start = self._tmd_stop
            stop = self._tmd_stop + self._jmd_c_len
            jmd_c = self._seq[start:stop]
        else:
            start = self._tmd_stop
            stop = self._tmd_stop + self._c_terminus_len
            part = self._seq[start:stop]
            jmd_c = part + ut.STR_AA_GAP * (self._jmd_c_len - len(part))  # Add "-" (gap) for missing AA in JMD
        return jmd_c


def create_parts(seq=None, tmd_start=None, tmd_stop=None, jmd_n_len=10, jmd_c_len=10):
    """Create JMD_N, TMD, and JMD_C using Parts object"""
    parts = Parts(seq=seq, tmd_start=tmd_start, tmd_stop=tmd_stop, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    jmd_n, tmd, jmd_c = parts.jmd_n, parts.tmd, parts.jmd_c
    return jmd_n, tmd, jmd_c


"""
This is a script for protein benchmark loading function. Please define new loading functions by their loaded data
by introducing a new data table in docs/source/index/tables_templates.rst.
"""
from typing import Optional, Literal, Union
import os
import pandas as pd
import numpy as np
import re
import warnings
from pandas import DataFrame
import aaanalysis.utils as ut

# Constants
FOLDER_BENCHMARKS = ut.FOLDER_DATA + "benchmarks" + ut.SEP
LIST_CANONICAL_AA = ['N', 'A', 'I', 'V', 'K', 'Q', 'R', 'M', 'H', 'F', 'E', 'D', 'C', 'G', 'L', 'T', 'S', 'Y', 'W', 'P']
LIST_NON_CANONICAL_OPTIONS = ["remove", "keep", "gap"]


# I Helper Functions
# TODO add df_cat_pc explaining PCs using loadings (if loadings ready)
# Check functions
def check_name_of_dataset(name="Overview", folder_in=None):
    """Check if name of dataset is valid"""
    if name == "Overview":
        return
    list_datasets = [x.split(".")[0] for x in os.listdir(folder_in)
                     if "." in x and not x.startswith(".")]
    if name not in list_datasets:
        list_aa = [x for x in list_datasets if 'AA' in x]
        list_seq = [x for x in list_datasets if 'SEQ' in x]
        list_dom = [x for x in list_datasets if 'DOM' in x]
        raise ValueError(f"'name' ({name}) is not valid. Chose one of the following:"
                         f"\n Amino acid datasets: {list_aa}"
                         f"\n Sequence datasets: {list_seq}"
                         f"\n Domain datasets: {list_dom}")


def check_min_max_val(min_len=None, max_len=None):
    """Check if min_val and max_val are valid and match"""
    ut.check_number_range(name="min_len", val=min_len, min_val=1, accept_none=True, just_int=True)
    ut.check_number_range(name="max_len", val=max_len, min_val=1, accept_none=True, just_int=True)
    if min_len is None or max_len is None:
        return
    if isinstance(min_len, int) and isinstance(max_len, int) and min_len > max_len:
        raise ValueError(f"'min_len' ({min_len}) should not be smaller than 'max_len' ({max_len})")


def check_non_canonical_aa(non_canonical_aa="remove"):
    """Check if non_canonical_aa is valid"""
    if non_canonical_aa not in LIST_NON_CANONICAL_OPTIONS:
        raise ValueError(f"'non_canonical_aa' ({non_canonical_aa}) should be on of following:"
                         f" {LIST_NON_CANONICAL_OPTIONS }")


def check_aa_window_size(aa_window_size=None):
    """Check if aa_window size is a positive odd integer"""
    if aa_window_size is None:
        return
    ut.check_number_range(name="aa_window_size", val=aa_window_size, min_val=1, just_int=True)
    if aa_window_size % 2 == 0:
        raise ValueError(f"'aa_window_size' ({aa_window_size}) must be an odd number.")


def post_check_df_seq(df_seq=None, n=None, name=None):
    """Check if length of df_seq is valid"""
    max_n = df_seq[ut.COL_LABEL].value_counts().min()
    warning_message = f"'n' ({n}) is too high since the smaller class for '{name}' contains {max_n} samples." \
                      f"\nThis maximum value depends on the filtering settings used."
    # Validation of sequence and domain datasets
    if n is not None and len(df_seq) != n*2:
        warnings.warn(warning_message)


# Helper functions
def _is_aa_level(name=None):
    return name.split("_")[0] == "AA"


def _adjust_non_canonical_aa(df=None, non_canonical_aa="remove"):
    """Adjust non-canonical amino acids"""
    if non_canonical_aa == "keep":
        return df
    # Get all non-canonical amino acids
    f = lambda x: set(str(x))
    vf = np.vectorize(f)
    char_seq = set().union(*vf(df.values).flatten())
    list_non_canonical_aa = [x for x in char_seq if x not in LIST_CANONICAL_AA]
    if non_canonical_aa == "remove":
        pattern = '|'.join(list_non_canonical_aa)  # Joining list into a single regex pattern
        df = df[~df[ut.COL_SEQ].str.contains(pattern, regex=True)]
    else:
        df[ut.COL_SEQ] = [re.sub(f'[{"".join(list_non_canonical_aa)}]', ut.STR_AA_GAP, x) for x in df[ut.COL_SEQ]]
    return df


def _get_aa_window(df_seq=None, aa_window_size=9):
    """Get amino acid windows from df_seq"""
    min_seq_len = df_seq[ut.COL_SEQ].apply(len).min()
    if df_seq[ut.COL_SEQ].apply(len).min() <= aa_window_size:
        raise ValueError(f"'aa_window_size' ({aa_window_size}) should be smaller than shortest sequence ({min_seq_len})")
    n_pre = n_post = int((aa_window_size - 1) / 2)  # checked to be an odd number before
    list_aa = []
    list_labels = []
    list_entries = []
    for entry, seq, labels in zip(df_seq[ut.COL_ENTRY], df_seq[ut.COL_SEQ], df_seq[ut.COL_LABEL]):
        for i in range(len(seq)):
            start_pre, end_pre = max(i - n_pre, 0), i
            start_post, end_post = i + 1, i + n_post + 1
            aa_window = seq[start_pre:end_pre] + seq[i] + seq[start_post:end_post]
            list_aa.append(aa_window)
            entry_pos = f"{entry}_pos{i}"
            list_entries.append(entry_pos)
        list_labels.extend(labels.split(","))
    df_seq = pd.DataFrame({ut.COL_ENTRY: list_entries, ut.COL_SEQ: list_aa, ut.COL_LABEL: list_labels})
    df_seq = df_seq[df_seq[ut.COL_SEQ].apply(len) == aa_window_size]    # Remove too short windows at sequence edges
    return df_seq


# II Main Functions
def load_dataset(name: str = "Overview",
                 n: Optional[int] = None,
                 random: bool = False,
                 non_canonical_aa: Literal["remove", "keep", "gap"] = "remove",
                 min_len: Optional[int] = None,
                 max_len: Optional[int] = None,
                 aa_window_size: Union[int, None] = 9,
                 ) -> DataFrame:
    """
    Load protein benchmarking datasets.

    The benchmarks are categorized into amino acid ('AA'), domain ('DOM'), and sequence ('SEQ') level datasets.
    By default, an overview table is provided (``name='Overview'``). For in-depth details, refer to [Breimann24a]_.

    Parameters
    ----------
    name : str, default='Overview'
        The name of the loaded dataset, from the 'Dataset' column in the overview table.
    n : int, optional
        Number of proteins per class, selected by index. If ``None``, the whole dataset will be returned.
    random : bool, default=False
        If True, ``n`` randomly selected proteins per class will be chosen.
    non_canonical_aa : {'remove', 'keep', 'gap'}, default='remove'
        Options for handling non-canonical amino acids:

        - ``remove``: Remove sequences containing non-canonical amino acids.
        - ``keep``: Don't remove sequences containing non-canonical amino acids.
        - ``gap``: Non-canonical amino acids are replaced by the gap symbol ('X').

    min_len : int, optional
        Minimum length of sequences for filtering.
    max_len : int, optional
        Maximum length of sequences for filtering.
    aa_window_size : int, default=9
        Length of amino acid window, only used for the amino acid dataset level (``name='AA_'``). Disabled if ``None``.

    Returns
    -------
    pandas.DataFrame
        A DataFrame of either the selected sequence dataset (``df_seq``) or
        overview on all benchmark datasets (``df_overview``).

    Notes
    -----
    ``df_seq`` includes these columns:

    - 'entry': Protein identifier, either the UniProt accession number or an id based on index.
    - 'sequence': Amino acid sequence.
    - 'label': Binary classification label (0 for negatives, 1 for positives).
    - 'tmd_start', 'tmd_stop': Start and stop positions of TMD (present only at the domain level).
    - 'jmd_n', 'tmd', 'jmd_c': Sequences for JMD_N, TMD, and JMD_C respectively.

    See Also
    --------
    * Overview of all benchmarks in :ref:`t1_overview_benchmarks`.
    * Step-by-step guide in the `Data Loading Tutorial <tutorial2a_data_loader.html>`_.

    Examples
    --------
    .. include:: examples/load_dataset.rst
    """
    check_name_of_dataset(name=name, folder_in=FOLDER_BENCHMARKS)
    ut.check_number_range(name="n", val=n, min_val=1, accept_none=True, just_int=True)
    check_non_canonical_aa(non_canonical_aa=non_canonical_aa)
    check_min_max_val(min_len=min_len, max_len=max_len)
    check_aa_window_size(aa_window_size=aa_window_size)
    # Load overview table
    if name == "Overview":
        return ut.read_excel_cached(FOLDER_BENCHMARKS + "Overview.xlsx").copy()
    df = ut.read_csv_cached(FOLDER_BENCHMARKS + name + ".tsv", sep="\t")
    # Filter data
    if min_len is not None:
        mask = [len(x) >= min_len for x in df[ut.COL_SEQ]]
        df = df[mask]
        if len(df) == 0:
            raise ValueError(f"'min_len' ({min_len}) is too high and removed all sequences.")
    if max_len is not None:
        mask = [len(x) <= max_len for x in df[ut.COL_SEQ]]
        df = df[mask]
        if len(df) == 0:
            raise ValueError(f"'max_len' ({max_len}) is too low and removed all sequences.")
    # Adjust non-canonical amino acid (keep, remove, or replace by gap)
    df_seq = _adjust_non_canonical_aa(df=df, non_canonical_aa=non_canonical_aa)
    # Adjust amino acid windows for 'AA' datasets
    if _is_aa_level(name=name):
        # Special case that unfiltered df_seq is returned
        if aa_window_size is None:
            return df_seq.reset_index(drop=True).copy()
        df_seq = _get_aa_window(df_seq=df_seq, aa_window_size=aa_window_size)
    # Select balanced groups
    if n is not None:
        labels = set(df_seq[ut.COL_LABEL])
        if random:
            df_seq = pd.concat([df_seq[df_seq[ut.COL_LABEL] == l].sample(n) for l in labels])
        else:
            df_seq = pd.concat([df_seq[df_seq[ut.COL_LABEL] == l].head(n) for l in labels])
    post_check_df_seq(df_seq=df_seq, n=n, name=name)
    # Adjust index and column values
    df_seq = df_seq.reset_index(drop=True)
    df_seq[ut.COL_LABEL] = df_seq[ut.COL_LABEL].astype(int)
    return df_seq.copy()




"""
Plotting utility functions for AAanalysis to create publication ready figures. Can
be used for any Python project independently of AAanalysis.
"""
from typing import Union
import seaborn as sns
import matplotlib as mpl
import matplotlib.pyplot as plt
import aaanalysis.utils as ut
import warnings


LIST_FONTS = ['Arial', 'Courier New', 'DejaVu Sans', 'Times New Roman', 'Verdana']


# I Helper functions
# Check plot_settings
def check_font(font="Arial"):
    if font not in LIST_FONTS:
        error_message = f"'font' ({font}) not in recommended fonts: {LIST_FONTS}. Set font manually by:" \
                        f"\n\tplt.rcParams['font.sans-serif'] = '{font}'"
        raise ValueError(error_message)


# Helper function
def set_tick_size(axis=None, major_size=None, minor_size=None):
    """Set tick size of the given axis."""
    plt.rcParams[f"{axis}tick.major.size"] = major_size
    plt.rcParams[f"{axis}tick.minor.size"] = minor_size


# II Main functions
def plot_settings(font_scale: Union[int, float] = 1,
                  font: str = "Arial",
                  weight_bold: bool = True,
                  adjust_only_font: bool = False,
                  adjust_further_elements: bool = True,
                  grid: bool = False,
                  grid_axis: str = "y",
                  no_ticks: bool = False,
                  short_ticks: bool = False,
                  no_ticks_x: bool = False,
                  short_ticks_x: bool = False,
                  no_ticks_y: bool = False,
                  short_ticks_y: bool = False,
                  show_options: bool = False
                  ) -> None:
    """
    Configure general plot settings.

    This function modifies the global settings of :mod:`matplotlib` and :mod:`seaborn` libraries.
    It adjusts font embedding for vector formats like PDF and SVG, ensuring compatibility and
    editability across various viewers and editing software.

    Parameters
    ----------
    font_scale : int or float, default=1
       Scaling factor to scale the size of font elements. Consistent with :func:`seaborn.set_context`.
    font : {'Arial', 'Courier New', 'DejaVu Sans', 'Times New Roman', 'Verdana'}, default='Arial'
       Name of text font. Common options are 'Arial' or 'DejaVu Sans' (Matplotlib default).
    weight_bold : bool, default=True
       If ``True``, font and line elements are bold.
    adjust_only_font : bool, default=False
       If ``True``, only the font style will be adjusted, leaving other elements unchanged.
    adjust_further_elements : bool, default=True
       If ``True``, makes additional visual and layout adjustments to the plot (errorbars, legend).
    grid : bool, default=False
       If ``True``, display the grid in plots.
    grid_axis : {'y', 'x', 'both'}, default='y'
       Choose the axis ('y', 'x', 'both') to apply the grid to.
    no_ticks : bool, default=False
       If ``True``, remove all tick marks on both x and y axes.
    short_ticks : bool, default=False
       If ``True``, display short tick marks on both x and y axes. Is ignored if ``no_ticks=True``.
    no_ticks_x : bool, default=False
       If ``True``, remove tick marks on the x-axis.
    short_ticks_x : bool, default=False
       If ``True``, display short tick marks on the x-axis. Is ignored if ``no_ticks=True``.
    no_ticks_y : bool, default=False
       If ``True``, remove tick marks on the y-axis.
    short_ticks_y : bool, default=False
       If ``True``, display short tick marks on the y-axis. Is ignored if ``no_ticks=True``.
    show_options : bool, default=False
       If ``True``, show all plot runtime configurations of matplotlib.

    Notes
    -----
    * ``grid_axis`` work only for axis with numerical values.

    See Also
    --------
    * More examples in `Plotting Prelude <plotting_prelude.html>`_.
    * :func:`seaborn.set_context`, where ``font_scale`` is utilized.
    * :data:`matplotlib.rcParams`, which manages the global settings in :mod:`matplotlib`.

    Examples
    --------
    .. include:: examples/plot_settings.rst
    """
    # Check input
    ut.check_number_range(name="font_scale", val=font_scale, min_val=0, just_int=False)
    check_font(font=font)
    ut.check_grid_axis(grid_axis=grid_axis)
    args_bool = {"weight_bold": weight_bold, "adjust_only_font": adjust_only_font,
                 "adjust_further_elements": adjust_further_elements, "grid": grid,
                 "short_ticks": short_ticks, "short_ticks_x": short_ticks_x, "short_ticks_y": short_ticks_y,
                 "no_ticks": no_ticks, "no_ticks_y": no_ticks_y, "no_ticks_x": no_ticks_x,
                 "show_options": show_options,}
    for key in args_bool:
        ut.check_bool(name=key, val=args_bool[key])

    # Warning
    if no_ticks and any([short_ticks, short_ticks_x, short_ticks_y]):
        warnings.warn("`no_ticks` is set to True, so 'short_ticks' parameters will be ignored.", UserWarning)
    if no_ticks_x and short_ticks_x:
        warnings.warn("`no_ticks_x` is set to True, so 'short_ticks_x' will be ignored.", UserWarning)
    if no_ticks_y and short_ticks_y:
        warnings.warn("`no_ticks_y` is set to True, so 'short_ticks_y' will be ignored.", UserWarning)

    # Print all plot settings/runtime configurations of matplotlib
    if show_options:
        print(plt.rcParams.keys)

    # Set all values to matplotlib default
    mpl.rcParams.update(mpl.rcParamsDefault)

    # Change only font style
    if adjust_only_font:
        mpl.rcParams['pdf.fonttype'] = 42  # Set embedded fonts in PDF via TrueType
        mpl.rcParams['ps.fonttype'] = 42
        mpl.rcParams['svg.fonttype'] = 'none'
        plt.rcParams["font.family"] = "sans-serif"
        plt.rcParams["font.sans-serif"] = font
        return

    # Apply all changes
    sns.set_context("talk", font_scale=font_scale)

    # Handle storing of vectorized figure format
    mpl.rcParams['pdf.fonttype'] = 42  # Set embedded fonts in PDF via TrueType
    mpl.rcParams['ps.fonttype'] = 42
    mpl.rcParams['svg.fonttype'] = 'none'

    # Font settings
    plt.rcParams["font.family"] = "sans-serif"
    plt.rcParams["font.sans-serif"] = font
    _weight = "bold" if weight_bold else "normal"
    font_settings = {'family': 'sans-serif', "weight": _weight}
    mpl.rc('font', **font_settings)

    # Grid
    plt.rcParams["axes.grid.axis"] = grid_axis
    plt.rcParams["axes.grid"] = grid
    plt.rcParams["grid.linewidth"] = 1 if weight_bold else 0.8
    plt.rcParams["axes.axisbelow"] = True

    # Adjust weight of text and lines
    if weight_bold:
        plt.rcParams["axes.labelweight"] = "bold"
        plt.rcParams["axes.titleweight"] = "bold"
    else:
        plt.rcParams["axes.labelweight"] = "normal"
        plt.rcParams["axes.titleweight"] = "normal"
        plt.rcParams["axes.linewidth"] = 1
        plt.rcParams["xtick.major.width"] = 0.8
        plt.rcParams["xtick.minor.width"] = 0.6
        plt.rcParams["ytick.major.width"] = 0.8
        plt.rcParams["ytick.minor.width"] = 0.6

    # Tick sizes
    default_major_size = 6
    default_minor_size = 4
    short_major_size = 3.5
    short_minor_size = 2
    no_major_size = 0
    no_minor_size = 0
    # Set x ticks
    if no_ticks or no_ticks_x:
        set_tick_size(axis="x", major_size=no_major_size, minor_size=no_minor_size)
    elif short_ticks or short_ticks_x:
        set_tick_size(axis="x", major_size=short_major_size, minor_size=short_minor_size)
    else:
        set_tick_size(axis="x", major_size=default_major_size, minor_size=default_minor_size)

    # Set y ticks
    if no_ticks or no_ticks_y:
        set_tick_size(axis="y", major_size=no_major_size, minor_size=no_minor_size)
    elif short_ticks or short_ticks_y:
        set_tick_size(axis="y", major_size=short_major_size, minor_size=short_minor_size)
    else:
        set_tick_size(axis="y", major_size=default_major_size, minor_size=default_minor_size)

    # Additional adjustments
    if adjust_further_elements:
        # Error bars
        plt.rcParams["errorbar.capsize"] = 10
        # Legend
        plt.rcParams["legend.frameon"] = False
        plt.rcParams["legend.fontsize"] = "medium"
        plt.rcParams["legend.loc"] = 'upper right'



"""This is a script for scikit-learn model-specific check functions"""
import inspect
from inspect import isclass

from ._utils import add_str

# Helper functions


# Main functions
def check_mode_class(model_class=None, str_add=None):
    """Check if the provided object is a class and callable, typically used for validating model classes."""
    # Check if model_class is actually a class and not an instance
    if not isclass(model_class):
        str_error = add_str(str_error=f"'model_class' ('{model_class}') is not a model class. "
                                      f"Please provide a valid model class.",
                            str_add=str_add)
        raise ValueError(str_error)
    # Check if model is callable
    if not callable(getattr(model_class, "__call__", None)):
        str_error = add_str(str_error=f"'model_class' ('{model_class}') is not a callable model.",
                            str_add=str_add)
        raise ValueError(str_error)


def check_model_kwargs(model_class=None, model_kwargs=None, name_model_class="model_class",
                       param_to_check=None, method_to_check=None, attribute_to_check=None,
                       random_state=None, str_add=None):
    """
    Check if the provided model class contains specific parameters and methods. Filters 'model_kwargs' to include only
    valid parameters for the model class.

    Parameters:
        model_class: The class of the model to check.
        model_kwargs: A dictionary of keyword arguments for the model.
        name_model_class: Name of model class for model class kwargs
        param_to_check: A specific parameter to check in the model class.
        method_to_check: A specific method to check in the model class.
        attribute_to_check: A specific attribute to check in model class
        random_state: random state
        str_add: additional error string

    Returns:
        model_kwargs: A filtered dictionary of model_kwargs containing only valid parameters for the model class.
    """
    model_kwargs = model_kwargs or {}
    if model_class is None:
        str_error = add_str(str_error=f"'{name_model_class}' must be provided.", str_add=str_add)
        raise ValueError(str_error)
    valid_args = list(inspect.signature(model_class).parameters.keys())
    # Check if 'param_to_check' is a parameter of the model
    if param_to_check is not None and param_to_check not in valid_args:
        str_error = add_str(str_error=f"'{param_to_check}' should be an argument in the given '{name_model_class}' ({model_class}).",
                            str_add=str_add)
        raise ValueError(str_error)
    # Check if 'method_to_check' is a method of the model
    if method_to_check is not None and not hasattr(model_class, method_to_check):
        str_error = add_str(str_error=f"'{method_to_check}' should be a method in the given '{name_model_class}' ({model_class}).",
                            str_add=str_add)
        raise ValueError(str_error)
    # Check if 'attribute_to_check' is an attribute of the model
    if attribute_to_check is not None and not hasattr(model_class, attribute_to_check):
        str_error = add_str(str_error=f"'{attribute_to_check}' should be an attribute in the given '{name_model_class}' ({model_class}).",
                            str_add=str_add)
        raise ValueError(str_error)
    # Check if model_kwargs contain invalid parameters for the model
    invalid_kwargs = [x for x in model_kwargs if x not in valid_args]
    if len(invalid_kwargs):
        str_error = add_str(str_error=f"'model_kwargs' (for '{model_class}') contains invalid arguments: {invalid_kwargs}",
                            str_add=str_add)
        raise ValueError(str_error)
    if "random_state" not in model_kwargs and "random_state" in valid_args:
        model_kwargs.update(dict(random_state=random_state))
    return model_kwargs





"""
This is a script for the frontend of the dPULearn class, used for deterministic Positive-Unlabeled (PU) Learning.
"""
from typing import Optional, Literal, Dict, Union, List, Tuple, Type
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

import aaanalysis.utils as ut
from ._backend.dpulearn.dpul_fit import (get_neg_via_distance, get_neg_via_pca)
from ._backend.dpulearn.dpul_eval import eval_identified_negatives
from ._backend.dpulearn.dpul_compare_sets_neg import compare_sets_negatives_

# Settings
LIST_METRICS = ['euclidean', 'manhattan', 'cosine']


# I Helper Functions
# Check functions
def check_metric(metric=None):
    """Validate provided metric"""
    if metric is not None and metric not in LIST_METRICS:
        raise ValueError(f"'metric' ({metric}) should be None or one of following: {LIST_METRICS}")


def check_n_unl_to_neg(labels=None, n_unl_to_neg=None, label_unl=None):
    """Validate that there are enough unlabeled samples in the dataset."""
    n_unl = np.sum(labels == label_unl)
    if n_unl < n_unl_to_neg:
        raise ValueError(f"Number of unlabeled labels ({n_unl}) must be higher than 'n_unl_to_neg' ({n_unl_to_neg})")


def check_n_components(n_components=1):
    """Check if n_components valid for sklearn PCA object"""
    try:
        # Check number of PCs
        if type(n_components) is int:
            ut.check_number_range(name="n_components", val=n_components, min_val=1, just_int=True)
        # Check percentage of covered explained variance
        else:
            ut.check_number_range(name="n_components", val=n_components, min_val=0, max_val=1.0, just_int=False,
                                  exclusive_limits=True)
    except ValueError:
        raise ValueError(f"'n_components' ({n_components}) should be either "
                         f"\n  an integer >= 1 (number of principal components) or"
                         f"\n  a float with 0.0 < 'n_components' < 1.0 (percentage of covered variance)")


def check_match_X_n_components(X=None, n_components=1):
    """Check if n_components matches to dimensions of X"""
    n_samples, n_features = X.shape
    if min(n_features, n_samples) <= n_components:
        raise ValueError(f"'n_components' ({n_components}) should be < min(n_features, n_samples) from 'X' ({n_features})")


def check_match_list_labels_df_seq(list_labels=None, df_seq=None):
    """Check if length of labels in list_labels and df_seq matches"""
    if df_seq is None:
        return None # Skip check
    n_samples = len(list_labels[0])
    if n_samples != len(df_seq):
        raise ValueError(f"Number of samples (n={n_samples}) in 'list_labels' does not match with "
                         f"samples in 'df_seq' (n={len(df_seq)})")

def check_match_X_X_neg(X=None, X_neg=None):
    """Check if number of features matches in both feature matrices"""
    if X_neg is None:
        return # Skip test
    n_features = X.shape[1]
    n_features_neg =  X_neg.shape[1]
    if n_features != n_features_neg:
        raise ValueError(f"'n_features' does not match between 'X' (n={n_features}) and 'X_neg' (n={n_features_neg})")


# II Main Functions
class dPULearn:
    """
    Deterministic Positive-Unlabeled Learning (**dPULearn**) class for identifying reliable negatives from unlabeled data [Breimann24c]_.

    dPULearn offers a deterministic approach to Positive-Unlabeled (PU) learning, featuring two distinct
    identification approaches:

    - **PCA-based identification**: This is the primary method where Principal Component Analysis (PCA) is utilized
      to reduce the dimensionality of the feature space. Based on the most informative principal components (PCs),
      the model iteratively identifies reliable negatives (labeled by 0) from the set of unlabeled samples (2).
      These reliable negatives are those that are most distant from the positive samples (1) in the feature space.
    - **Distance-based identification**: As a simple alternative, reliable negatives can also be identified using
      similarity measures like ``euclidean``, ``manhattan``, or ``cosine`` distance.

    Attributes
    ----------
    labels_ : array-like, shape (n_samples,)
        New dataset labels of samples in ``X`` with identified negative samples labeled by 0.
    df_pu_ : pd.DataFrame, shape (n_samples, pca_features)
        A DataFrame with the PCA-transformed features of 'X' containing the following groups of columns:

        - 'selection_via': Column indicating how reliable negatives were identified (either giving the distance metric
           or the i-th PC based on which the respective sample was selected).
        - 'PCi': Value columns for the i-th principal component (PC).
        - 'PCi_abs_dif': Absolute difference columns for each PC, representing the absolute deviation of each sample
          from the mean of positives.

        For distance-based identification, 'PCi' columns are replaced with the results for the selected metric.


    """
    def __init__(self,
                 model_kwargs: Optional[dict] = None,
                 verbose: bool = True,
                 random_state: Optional[str] = None,
                 ):
        """
        Parameters
        ----------
        model_kwargs : dict, optional
            Additional keyword arguments for Principal Component Analysis (PCA) model.
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        Notes
        -----
        * All attributes are set during fitting via the :meth:`dPULearn.fit` method and can be directly accessed.
        * For a detailed discussion on Positive-Unlabeled (PU) learning, its challenges, and evaluation strategies,
          refer to the PU Learning section in the Usage Principles documentation: `usage_principles/pu_learning`.

        See Also
        --------
        * :class:`dPULearnPlot`: the respective plotting class.
        * :func:`sklearn.decomposition.PCA` for details on principal component analysis.
        """
        # Global parameters
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        # Model parameters
        model_kwargs = ut.check_model_kwargs(model_class=PCA,
                                             model_kwargs=model_kwargs,
                                             random_state=random_state)
        # Internal attributes
        self._verbose = verbose
        self._random_state = random_state
        self._model_kwargs = model_kwargs
        # Output parameters (will be set during model fitting)
        self.labels_ = None
        self.df_pu_ = None

    # Main method
    def fit(self,
            X: ut.ArrayLike2D,
            labels: ut.ArrayLike1D = None,
            n_unl_to_neg: int = None,
            metric: Optional[Literal["euclidean", "manhattan", "cosine"]] = None,
            n_components: Union[float, int] = 0.80,
            ) -> "dPULearn":
        """
        Fit the dPULearn model to identify reliable negative samples (labeled by 0) from unlabeled samples (2)
        based on the distance to positive samples (1).

        Use the ``dPUlearn.labels_`` attribute to retrieve the output labels of samples in ``X``
        including identified negatives.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        labels : array-like, shape (n_samples,)
            Dataset labels of samples in ``X``. Should be either 1 (positive) or 2 (unlabeled).
        n_unl_to_neg : int, default=1
            Number of negative samples (0) to be reliably identified from unlabeled samples (2).
            Should be < n unlabeled samples.
        metric : str or None, optional
            The distance metric to use. If ``None``, PCA-based identification is performed. For distance-based
            identification one of the following measures can be selected:

            - ``euclidean``: Euclidean distance (minimum)
            - ``manhattan``: Manhattan distance (minimum)
            - ``cosine``: Cosine distance (minimum)

        n_components : int or float, default=0.80
            Number of principal components (a) or the percentage of total variance to be covered (b) when PCA is applied.

            - In case (a): it should be an integer >= 1.
            - In case (b): it should be a float with  0.0 < ``n_components`` < 1.0.

        Returns
        -------
        dPULearn
            The fitted instance of the dPULearn class, allowing direct attribute access.

        Notes
        -----
        * If a distance metric is specified, dPUlearn performs distance-based instead of PCA-based identification.
        * When selecting a distance metric for distance-based identification, consider the dimensionality of the
          feature space, determined by the ratio of the number of features (n_features) to the number of samples
          (n_samples) in `X`. In a low-dimensional space, there are fewer features than samples (n_features < n_samples),
          whereas a high-dimensional space has significantly more features than samples (n_features >> n_samples).
          The choice of metric depends on the specific application, with the following general guidelines:

          - ``euclidean``: Effective in low-dimensional spaces or when direct distances are meaningful.
          - ``manhattan``: Useful when differences along individual dimensions are important, or in the presence of outliers.
          - ``cosine``: Recommended for high-dimensional spaces (e.g., n_features >> n_samples), as it evaluates
            the direction of feature vectors between data points rather than the magnitude of their differences.

        See Also
        --------
        * See `scikit-learn <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics.pairwise>`_
          for details the three different pairwise distance measures.
        * See [Hastie09]_ for a detailed explanation on feature space and high-dimensional problems.

        Examples
        --------
        .. include:: examples/dpul_fit.rst
        """
        # Check input
        X= ut.check_X(X=X)
        ut.check_labels(labels=labels, vals_requiered=[1, 2], allow_other_vals=False)
        ut.check_number_range(name="n_unl_to_neg", val=n_unl_to_neg, min_val=1, just_int=True)
        check_n_unl_to_neg(labels=labels, n_unl_to_neg=n_unl_to_neg, label_unl=2)
        check_metric(metric=metric)
        check_n_components(n_components=n_components)
        ut.check_match_X_labels(X=X, labels=labels)
        check_match_X_n_components(X=X, n_components=n_components)
        # Compute average distance for threshold-based filtering (Yang et al., 2012, 2014; Nan et al. 2017)
        args = dict(X=X, labels=labels, n_unl_to_neg=n_unl_to_neg, label_neg=0)
        if metric is not None:
            new_labels, df_pu = get_neg_via_distance(**args, metric=metric)
        # Identify most far away negatives in PCA compressed feature space
        else:
            new_labels, df_pu = get_neg_via_pca(**args, n_components=n_components, **self._model_kwargs)
        # Set new labels
        self.labels_ = np.asarray(new_labels)
        self.df_pu_ = df_pu
        return self

    @staticmethod
    def eval(X: ut.ArrayLike2D,
             list_labels: ut.ArrayLike2D = None,
             names_datasets: Optional[List[str]] = None,
             X_neg: Optional[ut.ArrayLike2D] = None,
             comp_kld: bool = False,
             ) -> pd.DataFrame:
        """
        Evaluates the quality of different sets of identified negatives.

        The quality is assessed regarding two quality groups:

        - **Homogeneity** within the reliably identified negatives (0)
        - **Dissimilarity** between the reliably identified negatives and the groups
          of positive samples ('pos'), unlabeled samples ('unl'), and a ground-truth negative
          ('neg') sample group if provided by ``X_neg``

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        list_labels : array-like, shape (n_datasets, n_samples)
            List of arrays with dataset labels for samples in ``X`` obtained by the :meth:`dPULearn.fit` method.
            Label values should be either 0 (identified negative), 1 (positive) or 2 (unlabeled).
        names_datasets : list, optional
            List of dataset names corresponding to ``list_labels``.
        X_neg : array-like, shape (n_samples_neg, n_features), optional
            Feature matrix where `n_samples_neg` is the number ground-truth negative samples
            and `n_features` is the number of features. Features must correspond to ``X``.
        comp_kld : bool, default=False
            Whether to compute Kullback-Leibler Divergence (KLD) to assess the distribution alignment between
            identified negatives and other data groups. Disable (``False``) if ``X`` is sparse or has low co-variance.

        Returns
        -------
        df_eval : pd.DataFrame
            Evaluation results for each set of identified negatives from ``list_labels``. For each set, statistical
            measures were averaged across all features.

        Notes
        -----
        ``df_eval`` includes the following columns:

        - 'name': Name of the dataset if ``names`` is provided (typically named by identification approach).
        - 'n_rel_neg': Number of identified negatives.
        - 'avg_std': Average standard deviation (STD) assessing homogeneity of identified negatives.
          Lower values indicate greater homogeneity.
        - 'avg_iqr': Average interquartile range (IQR) assessing homogeneity of identified negatives.
          Lower values suggest greater homogeneity.
        - 'avg_abs_auc_DATASET': Average absolute area under the curve (AUC) assessing the dissimilarity between the
          set of identified negatives with other groups (positives, unlabeled, ground-truth negatives).
          Separate columns are provided for each comparison. Higher values indicate greater dissimilarity.
        - 'avg_kld_DATASET': Average Kullback-Leibler Divergence (KLD) assessing the dissimilarity of distributions
          between the set of identified negatives and the other groups. Higher values indicate greater dissimilarity.
          These columns are omitted if ``kld`` is set to ``False``.

        See Also
        --------
        * :meth:`dPULearnPlot.eval`: the respective plotting method.
        * :ref:`usage_principles_pu_learning` for details on different evaluation strategies.

        Examples
        --------
        .. include:: examples/dpul_eval.rst
        """
        # Check input
        X= ut.check_X(X=X)
        X_neg = ut.check_X(X=X_neg, X_name="X_neg", accept_none=True, min_n_samples=2)
        ut.check_bool(name="comp_kld", val=comp_kld)
        list_labels = ut.check_array_like(name="list_labels", val=list_labels, ensure_2d=True, convert_2d=True)
        names_datasets = ut.check_list_like(name="names_datasets", val=names_datasets, accept_none=True, accept_str=True,
                                            check_all_str_or_convertible=True)
        ut.check_match_X_list_labels(X=X, list_labels=list_labels, check_variability=comp_kld, vals_requiered=[0])
        ut.check_match_list_labels_names_datasets(list_labels=list_labels, names_datasets=names_datasets)
        check_match_X_X_neg(X=X, X_neg=X_neg)
        # Evaluation for homogeneity within negatives and alignment of distribution with other datasets
        df_eval = eval_identified_negatives(X=X, list_labels=list_labels, names_datasets=names_datasets,
                                            X_neg=X_neg, comp_kld=comp_kld)
        return df_eval

    @staticmethod
    def compare_sets_negatives(list_labels: ut.ArrayLike1D = None,
                               names_datasets: Optional[List[str]] = None,
                               df_seq: Optional[pd.DataFrame] = None,
                               remove_non_neg : bool = True,
                               return_upset_data: bool = False
                               ) -> pd.DataFrame:
        """
        Create DataFrame for comparing sets of identified negatives.

        Optionally, data format can be created for Upset Plots, which are useful for visualizing the intersection
        and unique elements across these sets.

        Parameters
        ----------
        list_labels : array-like, shape (n_datasets,)
            List of dataset labels for samples in ``X`` obtained by the :meth:`dPULearn.fit` method.
            Label values should be either 0 (identified negative), 1 (positive) or 2 (unlabeled). Must contain 0.
        names_datasets : list, optional
            List of dataset names corresponding to ``list_labels``.
        df_seq : pd.DataFrame, shape (n_samples, n_seq_info), optional
            DataFrame with sequence information for entries corresponding to 'labels' of ``list_labels``.
        remove_non_neg : bool, default=True
            If ``True``, all rows are removed that do not contain identified negatives in any provided dataset.
        return_upset_data : bool, default=False
            Whether to return a DataFrame for Upset Plot (if ``True``) or for a general comparison of sets of negatives.

        Returns
        -------
        pd.DataFrame or pd.Series
            - If ``return_upset_data=False`` (default):
              Returns a pd.DataFrame (`df_neg_comp`) that combines ``df_seq`` (if provided) with a comparison of the
              negative sets for a general analysis.
            - If ``return_upset_data=True``:
              Returns a pd.Series DataFrame (`upset_data`) formatted for generating  Upset Plots, containing group
              size information for the intersection and unique elements across the label sets.

        See Also
        --------
        * :meth:`dPULearn.fit` for details on how labels are generated.
        * :meth:`SequenceFeature.get_df_parts` for details on format of ``df_seq``.
        * Upset Plot documentation: :func:`upsetplot.plot`.

        Examples
        --------
        .. include:: examples/dpul_compare_sets_negatives.rst
        """
        # Check input
        list_labels = ut.check_array_like(name="list_labels", val=list_labels,
                                          ensure_2d=True, convert_2d=True)
        names_datasets = ut.check_list_like(name="names_datasets", val=names_datasets, accept_none=True,
                                            accept_str=True, check_all_str_or_convertible=True)
        ut.check_df_seq(df_seq=df_seq, accept_none=True)
        ut.check_bool(name="return_upset_data", val=return_upset_data)
        ut.check_match_list_labels_names_datasets(list_labels=list_labels, names_datasets=names_datasets)
        check_match_list_labels_df_seq(list_labels=list_labels, df_seq=df_seq)
        # Comparison of identified sets of negatives
        args = dict(list_labels=list_labels, names=names_datasets,
                    df_seq=df_seq, return_upset_data=return_upset_data, remove_non_neg=remove_non_neg)
        if return_upset_data:
            upset_data = compare_sets_negatives_(**args)
            return upset_data
        df_neg_comp = compare_sets_negatives_(**args)
        return df_neg_comp



"""
This is a script for the backend of the dPULearnPlot methods.
"""
import time
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

import aaanalysis.utils as ut


# I Helper Functions
def _plot_neg_homogeneity(ax=None, df_eval=None, col=None, colors=None, val_name="Value"):
    """Plot homogeneity among identified negatives"""
    sns.barplot(ax=ax, data=df_eval, y=ut.COL_NAME, x=col, color=colors[0])
    # Adding the numbers to the end of the bars
    for p in ax.patches:
        val = p.get_width()
        val = '{:.3f}'.format(val) if val < 1 else int(val)
        ax.text(p.get_width() * 0.97, p.get_y() + p.get_height() / 2, val, ha='right', va='center')
    # Customize subplots
    ax.set_ylabel("")
    ax.set_xlabel(val_name)
    # Adjust spines
    ax = ut.adjust_spine_to_middle(ax=ax)
    ut.x_ticks_0(ax=ax)
    ax.tick_params(axis='y', which='both', left=False)


def _plot_dist_dissimilarity(ax=None, df_eval=None, cols=None, colors=None, val_name="Value"):
    """Plot how strong the distributions of identified negatives and the other groups differ"""
    df_auc = df_eval[[ut.COL_NAME] + cols]
    df_auc = df_auc.melt(id_vars=ut.COL_NAME, value_vars=cols, var_name="Reference", value_name=val_name)
    dict_color = dict(zip(cols, colors[1:len(cols) + 1]))
    ax = sns.barplot(ax=ax, data=df_auc, y=ut.COL_NAME, x=val_name,
                     palette=dict_color, hue="Reference", legend=False, edgecolor="white")
    # Customize subplots
    ax.set_ylabel("")
    ax.set_xlabel(val_name)
    # Adjust spines
    ax = ut.adjust_spine_to_middle(ax=ax)
    ut.x_ticks_0(ax=ax)
    ax.tick_params(axis='y', which='both', left=False)
    return ax


# II Main Functions
def plot_eval(df_eval=None, figsize=None, dict_xlims=None, colors=None, legend=True, legend_y=-0.175):
    """Plot evaluation of AAclust clustering results"""
    cols_eval = [x for x in ut.COLS_EVAL_DPULEARN if x in list(df_eval)]
    cols_homogeneity = cols_eval[1:3]
    cols_auc = [c for c in cols_eval if "AUC" in c]
    cols_kld = [c for c in cols_eval if "KLD" in c]
    kld_in = len(cols_kld) != 0
    n_colss = 3 if not kld_in else 4
    fig, axes = plt.subplots(1, n_colss, sharey=True, figsize=figsize)
    # Show Homogeneity
    for i, col in enumerate(cols_homogeneity):
        ax = axes[i]
        _plot_neg_homogeneity(ax=ax, df_eval=df_eval, col=col, colors=colors, val_name=col.replace("_", " "))
        if i == 0:
            ax.set_title("Homogeneity", weight="bold", ha="left")
    # Plot distribution dissimilarity
    ax = _plot_dist_dissimilarity(ax=axes[2], df_eval=df_eval, cols=cols_auc, colors=colors,  val_name="avg AUC")
    if kld_in:
        _plot_dist_dissimilarity(ax=axes[3], df_eval=df_eval, cols=cols_kld, colors=colors, val_name="avg KLD")
        ax.set_title("Dissimilarity", weight="bold", ha="left")
    else:
        ax.set_title("Dissimilarity", weight="bold")
    if legend:
        labels = [x.split("_")[3].capitalize() for x in cols_auc]
        dict_color = dict(zip(labels, colors[1:len(labels) + 1]))
        ut.plot_legend_(ax=ax, dict_color=dict_color,
                        title="Reference datasets", y=legend_y, handletextpad=0.1)
    # Adjust plot
    if dict_xlims is not None:
        for i in dict_xlims:
            # Check that KLD axis are only set if valid
            if i <= n_colss:
                axes[i].set_xlim(dict_xlims[i])
    plt.tight_layout()
    plt.subplots_adjust(wspace=0.25, hspace=0)
    return fig, axes


def plot_pca(df_pu=None, labels=None, figsize=(6, 6),
             pc_x=1, pc_y=2, show_pos_mean_x=False, show_pos_mean_y=False,
             names=None, colors=None,
             legend=True, legend_y=-0.15, args_scatter=None):
    """Generates a PCA plot based on provided parameters."""
    plt.figure(figsize=figsize)
    cols_pc = [x for x in list(df_pu) if "abs" not in x and "PC" in x]
    label_x = [x for x in cols_pc if f'PC{pc_x}' in x][0]
    label_y = [y for y in cols_pc if f'PC{pc_y}' in y][0]
    # Map colors to labels
    dict_color = {label: color for label, color in zip(sorted(set(labels)), colors)}
    fs = ut.plot_gco(option="font.size")
    # Plotting
    for label in reversed(sorted((set(labels)))):
        subset = df_pu[labels == label]
        plt.scatter(subset[label_x], subset[label_y], color=dict_color[label], label=label, **args_scatter)
    # Handling mean lines for positive samples
    if show_pos_mean_x or show_pos_mean_y:
        pos_samples = df_pu[labels == 1]  # Assuming label '1' is for positive samples
        mean_x = pos_samples[label_x].mean() if show_pos_mean_x else None
        mean_y = pos_samples[label_y].mean() if show_pos_mean_y else None
        lw = ut.plot_gco(option="lines.linewidth")
        args = dict(ha='right', fontsize=fs, color=colors[1])
        if mean_x is not None:
            str_mean_x = fr"$\bar{{x}}_{{\text{{{label_x.split(' ')[0]}}}}}$"
            plt.axvline(mean_x, color='black', linestyle='--', linewidth=lw)
            plt.text(mean_x, plt.gca().get_ylim()[1], str_mean_x,  va="top", **args)

        if mean_y is not None:
            str_mean_y = fr"$\bar{{x}}_{{\text{{{label_y.split(' ')[0]}}}}}$"
            plt.axhline(mean_y, color='black', linestyle='--', linewidth=lw)
            plt.text(plt.gca().get_xlim()[1], mean_y, str_mean_y, va="bottom", **args)

    # Legend settings
    if legend:
        dict_color = {label: color for label, color in zip(names, colors)}
        ut.plot_legend_(ax=plt.gca(), dict_color=dict_color, title="Datasets",
                        n_cols=1, y=legend_y, handletextpad=0.4, fontsize=fs, fontsize_title=fs - 1, weight_title="bold")
    # Labels
    plt.xlabel(label_x)
    plt.ylabel(label_y)
    sns.despine()
    plt.tight_layout()
    ax = plt.gca()
    return ax


"""
This is a script for displaying pd.DataFrames as HTML output for jupyter notebooks.
"""
from typing import Optional, Union
import pandas as pd
from IPython.display import display, HTML

from aaanalysis import utils as ut

# Helper functions
def _adjust_df(df=None, char_limit = 50):
    df = df.copy()
    list_index = df.index
    # Adjust index if it consists solely of integers
    if all(isinstance(i, int) for i in list_index):
        df.index = [x + 1 for x in df.index]
    # Function to truncate strings longer than char_limit
    def truncate_string(s):
        return str(s)[:int(char_limit/2)] + '...' + str(s)[-int(char_limit/2):] if len(str(s)) > char_limit else s
    # Apply truncation to each cell in the DataFrame
    if char_limit is not None:
        for col in df.columns:
            df[col] = df[col].apply(lambda x: truncate_string(x) if isinstance(x, str) else x)
    return df


def _check_show(name="row_to_show", val=None, df=None):
    """Check if valid string or int"""
    if val is None:
        return None # Skip test
    rows_or_columns = list(df.T) if "row" in name else list(df)
    n = len(rows_or_columns)
    str_row_or_column = "row" if "row" in name else "column"
    if isinstance(val, str):
        ut.check_str(name=name, val=val, accept_none=True)
        if val not in rows_or_columns:
            raise ValueError(f"'{name}' ('{val}') should be one of following: {rows_or_columns}")
    elif isinstance(val, int):
        ut.check_number_range(name=name, val=val, accept_none=True, min_val=0, max_val=n, just_int=True)
    else:
        raise ValueError(f"'{name}' ('{val}') should be int (<{n}) or one of following {str_row_or_column} names: {rows_or_columns}")

def _select_row(df=None, row_to_show=None):
    """Select row"""
    if row_to_show is not None:
        if isinstance(row_to_show, int):
            df = df.iloc[[row_to_show]]
        elif isinstance(row_to_show, str):
            df = df.loc[[row_to_show]]
    return df

def _select_col(df=None, col_to_show=None):
    """Select column"""
    if col_to_show is not None:
        if isinstance(col_to_show, int):
            df = df.iloc[:, [col_to_show]]
        elif isinstance(col_to_show, str):
            df = df[[col_to_show]]
    return df


# Main functions
def display_df(df: pd.DataFrame = None,
               max_width_pct: int = 100,
               max_height: int = 300,
               char_limit: int = 30,
               show_shape=False,
               n_rows: Optional[int] = None,
               n_cols: Optional[int] = None,
               row_to_show: Optional[Union[int, str]] = None,
               col_to_show: Optional[Union[int, str]] = None,
               ):
    """
    Display DataFrame with specific style as HTML output for jupyter notebooks.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to be displayed as HTML output.
    max_width_pct: int, default=100
        Maximum width in percentage of main page for table.
    max_height : int, default=300
       Maximum height in pixels of table.
    char_limit : int, default=30
        Maximum number of characters to display in a cell.
    show_shape : bool, default=False
        If ``True``, shape of ``df`` is printed.
    n_rows : int, optional
        Display only the first n rows. If negative, last n rows will be shown.
    n_cols : int, optional
        Display only the first n columns. If negative, last n columns will be shown.
    row_to_show : int or str, optional
        Display only the specified row.
    col_to_show : int or str, optional
        Display only the specified column.

    Examples
    --------
    .. include:: examples/display_df.rst
    """
    # Check input
    ut.check_df(name="df", df=df, accept_none=False)
    ut.check_number_range(name="max_width_pct", val=max_width_pct, min_val=1, max_val=100, accept_none=False, just_int=True)
    ut.check_number_range(name="max_height", val=max_height, min_val=1, accept_none=False, just_int=True)
    ut.check_number_range(name="char_limit", val=char_limit, min_val=1, accept_none=True, just_int=True)
    n_rows_, n_cols_ = len(df), len(df.T)
    ut.check_number_range(name="n_rows", val=n_rows, min_val=-n_rows_, max_val=n_rows_, accept_none=True, just_int=True)
    ut.check_number_range(name="n_cols", val=n_cols, min_val=-n_cols_, max_val=n_cols_, accept_none=True, just_int=True)
    _check_show(name="show_only_col", val=col_to_show, df=df)
    _check_show(name="show_only_row", val=row_to_show, df=df)
    # Show shape before filtering
    if show_shape:
        print(f"DataFrame shape: {df.shape}")
    # Filtering
    df = df.copy()
    df = _select_col(df=df, col_to_show=col_to_show)
    df = _select_row(df=df, row_to_show=row_to_show)
    if row_to_show is None and n_rows is not None:
        if n_rows > 0:
            df = df.head(n_rows)
        else:
            df = df.tail(abs(n_rows))
    if col_to_show is None and n_cols is not None:
        if n_cols > 0:
            df = df.T.head(n_cols).T
        else:
            df = df.T.tail(abs(n_cols)).T
    # Style dataframe
    df = _adjust_df(df=df, char_limit=char_limit)
    styled_df = (
        df.style
        .set_table_attributes(f"style='display:block; max-height: {max_height}px; max-width: {max_width_pct}%; overflow-x: auto; overflow-y: auto;'")
        .set_table_styles([
            # Explicitly set background and text color for headers
            {'selector': 'thead th', 'props': [('background-color', 'white'), ('color', 'black')]},
             # Style for odd and even rows
            {'selector': 'tbody tr:nth-child(odd)', 'props': [('background-color', '#f2f2f2')]},
            {'selector': 'tbody tr:nth-child(even)', 'props': [('background-color', 'white')]},
            # General styling for table cells
            {'selector': 'th, td', 'props': [('padding', '5px'), ('white-space', 'nowrap')]},
        ])
    )
    display(HTML(styled_df.to_html()))
